import math
import re

#########################################
# define patterns to be found in string #
#########################################


# used in card_split
celmat = re.compile(
    r"(?P<cel>^ *\d+) +(?P<scnd>(\d+|like))", re.I
)  # identify first two entries on cell card (cell name, material)
grlnumber = re.compile(
    r"[-+]?(\d+\.\d+|\.\d+|\d+\.?)(e[+-]\d+)?", re.I
)  # identify a general number form signed integer, float or exponential
param = re.compile(
    r"((^|\n {5})[\(\):\-\+\d+\.\# ]*)([\*a-z])", re.I
)  # identity begining of the paramter part of the cell card
likebut = re.compile(r"but", re.I)  # identify likebut  card
trans = re.compile(r"trcl|fill= *\d+[ c\$\n]*\(,", re.I)  # identify tranformed card
unsignedint = re.compile(r"\d+")

# user in get_stat function
reword = re.compile(r"(\d+|\(|\)|:|\#)")  # word identification in cell line
compcell = re.compile(r"\#\d")  # identify hashcell complement operator


# used in Complementary operator function
number = re.compile(r"(?P<number>[-+]?\d+)")  # signed (+-) (or not) numbers
# leftp=re.compile(r"^ *(?P<left>[-\d\(\#])",re.M)                               # identify first valid character
leftp = re.compile(r"(?P<left>[-+\d\(\#])", re.I)  # identify first valid character
rightp = re.compile(r"(?P<right>[ c\$\n]*$)", re.I)  # identify last valid character
# interblk=re.compile(r"(?P<previous>\d)(?P<next>(( +| *(\$)?\n(C\n)* *)[-+]?\d))") # two numbers separated by blank (or newline or comments)
# intercls=re.compile(r"(?P<previous>\))(?P<next>(( *| *(\$)?\n(C\n)* *)[-+]?\d))") # closed parenthesis followed by number
# interopn=re.compile(r"(?P<previous>\d)(?P<next>(( *| *(\$)?\n(C\n)* *)\())")   # number followed by opened parenthesis
# intercop=re.compile(r"(?P<previous>\))(?P<next>(( *| *(\$)?\n(C\n)* *)\())")   # closed parenthesis followed by opened parenthesis
interblk = re.compile(
    r"(?P<previous>\d)(?P<next>(( +| *((\n *)?\$|\nC)*\n *)[-+]?\d))"
)  # two numbers separated by blank (or newline or comments)
intercls = re.compile(
    r"(?P<previous>\))(?P<next>(( *| *((\n *)?\$|\nC)*\n *)[-+]?\d))"
)  # closed parenthesis followed by number
interopn = re.compile(r"(?P<previous>\d)(?P<next>(( *| *((\n *)?\$|\nC)*\n *)\())")  # number followed by opened parenthesis
intercop = re.compile(
    r"(?P<previous>\))(?P<next>(( *| *((\n *)?\$|\nC)*\n *)\())"
)  # closed parenthesis followed by opened parenthesis
colonamp = re.compile(r"[:&]")  # colon or amperserand

# used for remove redundant parenthesis function
mostinner = re.compile(r"\([^\(^\)]*\)")  # identify most inner parentheses
bracketsemi = re.compile(r"[\]\[;]")  # square bracket or semicolon
blnkline = re.compile(r"^ *\n", re.M)  # identify blank line
contline = re.compile(r"\n {0,4}(?P<start>[^c^ ])", re.I)  # identify character other than 'C' in fisrt 5 columns
comdollar = re.compile(r"\n(?P<blnk> *)\$")  # identify dollar on 'blank line'
startgeom = re.compile(r"(?P<previous>^ *)(?P<start>[\-\+\d])")  # identify beginning of the geomtric part
endgeom = re.compile(r"(?P<last>\d)(?P<next> *((\n *)?\$|\nc)?(\n *)?$)", re.I)  # identify end of the geomtric part
# endgeom=re.compile(r"(?P<last>\d)(?P<next> *(\$|\nc)?(\n *)?$)",re.I)                      # identify end of the geomtric part

# other
rehash = re.compile(r"# *(\d+|\()")  # find beginning of complementary operator (both cell and surf)
parent = re.compile(r"[\(|\)]")  # position of open and close parenthesis (get_hashcell)
gline = re.compile(
    r"(^ ?[\(\):\-\+\d+\.\# ]+|\n {5}[\(\):\-\+\d+\.\# ]+)", re.I
)  # valid geometric part of the line       (remove/restore_comments)
comments = re.compile(r"((\n *)?\$|\n *c)", re.I)  # begining of comment part               (remove/restore_comments)
# comments=re.compile(r"\$|\n *c",re.I)                               # begining of comment part               (remove/restore_comments)
celtrsf = re.compile(r"TRCL *= *", re.I)
celuniverse = re.compile(r"U *= *", re.I)
celfill = re.compile(r"FILL *= *", re.I)
trfnumber = re.compile(r"([-+]?(\d+\.\d+|\.\d+|\d+\.?)(e[+-]\d+)?)|\)", re.I)  # search for general number or close bracket )
likemat = re.compile(r"MAT *= *(?P<mat>\d+)", re.I)  # identify material value on like but card
dollar = re.compile(r"\$.*\n", re.I)


############################################################
# Auxiliary functions used in regular expresion functions  #
############################################################
def remove_dollar(string):
    m = dollar.search(string)
    while m:
        string = string[0 : m.start()] + " " + string[m.end() :]
        m = dollar.search(string)
    return string


def redundant(m, geom):
    """check if the inner parentheses are redundant"""
    term = m.group()

    # Find first valid character at the left of the  parenthese
    hashsmb = False
    leftOK = True
    left = m.start() - 1
    while left > -1:
        if geom[left] in ("\n", "C", "$", " "):
            left -= 1
        else:
            if geom[left] not in ("(", ":"):
                leftOK = False
            if geom[left] == "#":
                hashsmb = True
            break

    # if hash symbol found means parentheses delimits complementary
    # cell defined with surface. Theses parentheses are not redundants
    if hashsmb:
        return False

    # check if no ':' (or) are inside the parenthese
    # if not, parentheses are redundants
    if term.find(":") == -1:
        return True

    # Find first valid character at the right of the  parenthese
    rightOK = True
    right = m.end()
    while right < len(geom):
        if geom[right] in ("\n", "C", "$", " "):
            right += 1
        else:
            if geom[right] not in (")", ":"):
                rightOK = False
            break

    # if parentheses are like:
    # {( or : } ( ....... ) {) or :}
    # parentheses are redundants

    if leftOK and rightOK:
        return True
    else:
        return False


# function used in Regular expresion sub function
# function user in complementary function
# change the sign of the number
def chgsign(m):
    num = m.group(0)
    if num[0] == "-":
        return num[1:]
    if num[0] == "+":
        return "-" + num[1:]
    else:
        return "-" + num


# function used in Regular expersion sub function
# function user in complementary function
# change ':' in ')('  and
#        '&' in ':'
def repl_inter_union(m):
    if m.group(0) == ":":
        return ")("
    else:
        return ":"


# function used in Regular expersion sub function
# function user in remove_redundant function
# restore curve parentheses and colon characters
def reverse_repl(m):
    symb = m.group(0)
    if symb == "[":
        return "("
    elif symb == "]":
        return ")"
    else:
        return ":"


############################################################


def complementary(ccell, outter=True):
    """return the complementary cell"""
    wrkcell = Cline(ccell.str)
    if wrkcell.str[-1] == "\n":
        wrkcell.str = wrkcell.str[:-1]

    # simplify comment in geometry string
    wrkcell.remove_comments()

    # insert external parenthesis
    wrkcell.str = re.sub(leftp, r"(\g<left>", wrkcell.str, count=1)
    wrkcell.str = re.sub(rightp, r")\g<right>", wrkcell.str, count=1)

    # insert '&' as intersection operator
    wrkcell.str = re.sub(
        interblk, r"\g<previous>&\g<next>", wrkcell.str
    )  # change intersection separate by blank space ie: "number number"
    wrkcell.str = re.sub(interblk, r"\g<previous>&\g<next>", wrkcell.str)  # 2nd pass intersection blank space (require 2 pass)
    wrkcell.str = re.sub(
        intercls, r"\g<previous>&\g<next>", wrkcell.str
    )  # change intersection close parenthesis ie: ") number"
    wrkcell.str = re.sub(
        interopn, r"\g<previous>&\g<next>", wrkcell.str
    )  # change intersection open  parenthesis ie: "number ("
    wrkcell.str = re.sub(
        intercop, r"\g<previous>&\g<next>", wrkcell.str
    )  # change intersection close-open  parenthesis ie: ") ("

    # invert operators
    # substitute colon by ')(' and  '&' by colon
    wrkcell.str = re.sub(colonamp, repl_inter_union, wrkcell.str)
    wrkcell.remove_redundant(remove_com=False)

    # Change signs
    wrkcell.str = re.sub(number, chgsign, wrkcell.str)

    # insert external parenthesis
    if outter:
        wrkcell.str = re.sub(leftp, r"(\g<left>", wrkcell.str, count=1)
        wrkcell.str = re.sub(rightp, r")\g<right>", wrkcell.str, count=1)

    # restore original comments
    wrkcell.restore_comments()
    return wrkcell.str


############################################################
class Cline:
    def __init__(self, line):
        self.str = line

    def copy(self):
        return Cline(self.str)

    def get_surfaces_numbers(self):
        s = set(unsignedint.findall(self.str))
        return tuple(map(int, s))

    def remove_multispace(self):
        self.str.strip()
        self.str = re.sub(r" +", " ", self.str)

    def remove_cr(self):
        self.str = re.sub(r"\n", " ", self.str)
        return

    def remove_comments(self, full=False):
        """Remove the text of the comment. The symbol 'C' or '$' is
        kept in the line if 'full' option is False (default)"""
        celltab = re.split(gline, self.str)
        cont = True
        while cont:
            try:
                celltab.remove("")
            except:
                cont = False

        self.__comtab__ = []
        for i, s in enumerate(celltab):
            c = comments.match(s)
            if c:
                if not full:
                    self.__comtab__.append(s)
                    celltab[i] = c.group()
                else:
                    celltab[i] = ""

        self.str = "".join(celltab)
        return

    def restore_comments(self):
        """Restore the text of the comment."""
        celltab = re.split(gline, self.str)
        cont = True
        while cont:
            try:
                celltab.remove("")
            except:
                cont = False

        j = 0
        for i, s in enumerate(celltab):
            c = comments.match(s)
            if c:
                celltab[i] = self.__comtab__[j]
                j += 1

        self.str = "".join(celltab)
        return

    def outer_terms(self):
        cgeom = Cline(self.str)

        cgeom.remove_comments(full=True)
        cgeom.remove_redundant()
        cgeom.remove_cr()
        geom = cgeom.str

        # Loop until no redundant parentheses are found
        cont = True

        while cont:
            # Loop over most inner parentheses
            pos = 0
            cont = False
            while True:
                m = mostinner.search(geom, pos)
                if not m:
                    break
                cont = True
                if redundant(m, geom):
                    # remove redundant parentheses
                    geom = geom[: m.start()] + " " + geom[m.start() + 1 : m.end() - 1] + " " + geom[m.end() :]
                else:
                    # replace no redundant parentheses by 0 and : by ;
                    zeros = "0" * (m.end() - m.start())
                    geom = geom[: m.start()] + zeros + geom[m.end() :]

                pos = m.end()

        if ":" in geom:
            terms = []
            pos = 0
            while True:
                newpos = geom.find(":", pos)
                if newpos == -1:
                    terms.append(cgeom.str[pos:].strip())
                    break
                terms.append(cgeom.str[pos:newpos].strip())
                pos = newpos + 1
            return (terms, "OR")
        else:
            terms = []
            pos = 0
            while True:
                m = number.search(geom, pos)
                if not m:
                    break
                terms.append(cgeom.str[m.start() : m.end()])
                pos = m.end()
            return (terms, "AND")

    def remove_redundant(self, remove_com=True, remopt="nochg"):
        """return cell without redundant parenthesis"""

        # simplify comment in geometry string
        if remove_com:
            self.remove_comments()
        geom = self.str

        if remopt == "nochg" and geom.find(")") == -1:
            self.removedp = None
            return

        porg = self.countP()
        # Loop until no redundant parentheses are found
        cont = True
        while cont:
            # Loop over most inner parentheses
            pos = 0
            cont = False
            while True:
                m = mostinner.search(geom, pos)
                if not m:
                    break
                cont = True
                if redundant(m, geom):
                    # remove redundant parentheses
                    geom = geom[: m.start()] + " " + geom[m.start() + 1 : m.end() - 1] + " " + geom[m.end() :]
                else:
                    # replace no redundant parentheses by [] and : by ;
                    term = geom[m.start() + 1 : m.end() - 1].replace(":", ";")
                    geom = geom[: m.start()] + "[" + term + "]" + geom[m.end() :]
                pos = m.end()

        # restore curved parenthesis and colon
        geom = re.sub(bracketsemi, reverse_repl, geom)

        # remove possible blank line
        geom = re.sub(blnkline, "", geom)

        # ensure 5 blanks continous line
        geom = re.sub(contline, r"\n     \g<start>", geom)

        if remopt != "all":
            # add parenthesis to set geom as MCNP complex cell
            if geom.find(":") == -1 and geom.find("#") == -1:
                geom = re.sub(startgeom, r"\g<previous>(\g<start>", geom)
                geom = re.sub(endgeom, r"\g<last>)\g<next>", geom)

        # restore original comments
        self.str = geom
        pmod = self.countP()
        if remove_com:
            self.restore_comments()

        # subtitute comment $ with  blank line
        self.str = re.sub(comdollar, r"\nC\g<blnk>", self.str)
        pdiff = [x - y for x, y in zip(pmod, porg)]
        self.removedp = pdiff
        return

    def get_hashcell(self, start=0):
        """get the complementary cell defined with surfaces combination"""
        count = 0
        for p in parent.finditer(self.str, start):
            if p.group() == "(":
                count += 1
            else:
                count -= 1
            if count == 0:
                end = p.end()
                cell = Cline(self.str[start + 1 : end])
                break
        return cell, end

    def replace(self, surf, new, pos=0):

        s1 = str(surf)
        s2 = str(new)
        m = unsignedint.search(self.str, pos)
        if not m:
            print("no number in line")
            return -1
        if surf == new:
            return m.start() + len(s2)
        elif m.group() == s1:
            self.str = self.str[0 : m.start()] + s2 + self.str[m.end() :]
            return m.start() + len(s2)
        else:
            print(f"number {surf} not found")
            return -1

    def countP(self):
        lp = self.str.count("(")
        rp = self.str.count(")")
        return (lp, rp)

    def SplitCell(self):
        self.remove_redundant(remopt="all")
        self.remove_comments()
        geom = self.str
        self.restore_comments()

        # insert '&' as intersection operator
        geom = re.sub(
            interblk, r"\g<previous>&\g<next>", geom
        )  # change intersection separate by blank space ie: "number number"
        geom = re.sub(interblk, r"\g<previous>&\g<next>", geom)  # 2nd pass intersection blank space (require 2 pass)
        geom = re.sub(intercls, r"\g<previous>&\g<next>", geom)  # change intersection close parenthesis ie: ") number"
        geom = re.sub(interopn, r"\g<previous>&\g<next>", geom)  # change intersection open  parenthesis ie: "number ("
        geom = re.sub(intercop, r"\g<previous>&\g<next>", geom)  # change intersection close-open  parenthesis ie: ") ("

        parts = []
        block = ""
        iopen = 0
        ctype = None

        for c in geom:
            if c == "\n":
                continue
            if c in [":", "&"]:
                if iopen == 0:
                    if ctype == None:
                        if c == "&":
                            ctype = "AND"
                        if c == ":":
                            ctype = "OR"
                    if ctype == "AND" and c == ":":
                        ctype = "OR"
                        break
            elif c == "(":
                iopen += 1
            elif c == ")":
                iopen -= 1

        for c in geom:
            if c == "\n":
                continue
            if c in [":", "&"]:
                if iopen == 0:
                    if ctype == "AND" and c == "&":
                        parts.append(block)
                        block = ""
                    if ctype == "OR" and c == ":":
                        parts.append(block)
                        block = ""
                else:
                    if c != "&":
                        block = block + c
            else:
                if c != "&":
                    block = block + c
                if c == "(":
                    iopen += 1
                elif c == ")":
                    iopen -= 1
        parts.append(block)
        return parts, ctype


############################################################
class CellCardString:

    def __init__(self, card):
        self.stat = {"word": None, "hashcell": None, "hashsurf": None, "hash": None}

        self.name = None
        self.MAT = None
        self.U = None
        self.FILL = None
        self.TR = None
        self.TRCL = None
        self.__card_split__(card)
        self.__get_data__()

    def __card_split__(self, cardin):
        """Split the card string in three parts :
           - headstr : string containing the cell name, mat number and density (if mat != 0) of the cell
           - geom    : Cline class containing the part of the geometric definition of the cell
           - param    : Cline class containing the cell parameters part
        hproc is true if the complementary operator of the cell can be substituted"""

        m = celmat.match(cardin)
        self.name = int(m.group("cel"))
        self.hproc = True
        self.likeCell = None
        if m.group("scnd").lower() == "like":
            self.headstr = cardin[: m.start("scnd")]
            s = likebut.search(cardin, m.end("scnd"))
            self.geom = Cline(cardin[m.start("scnd") : s.end()])
            self.parm = Cline(cardin[s.end() :])
            self.hproc = False
            mc = unsignedint.search(self.geom.str)
            self.likeCell = int(mc.group())
            ml = likemat.search(self.parm.str)
            if ml:
                self.MAT = int(ml.group("mat"))
        elif m.group("scnd") == "0":
            cstart = m.end("scnd")
            self.MAT = 0
        else:
            self.MAT = int(m.group("scnd"))
            p = grlnumber.search(cardin, m.end("scnd"))
            cstart = p.end()

        if self.hproc:
            self.headstr = cardin[:cstart]
            cellcard = Cline(cardin[cstart:])
            cellcard.remove_comments()
            m = param.search(cellcard.str)
            if m:
                linecut = cellcard.str[: m.end(1)].count("\n")
            else:
                linecut = cellcard.str.count("\n")

            cellcard.restore_comments()

            # look for the last line geometry string
            if linecut != 0:
                pos = 0
                c = 0
                while c != linecut:
                    pos = cellcard.str.find("\n", pos)
                    c += 1
                m = param.search(cellcard.str, pos)

            if m:
                start = m.end(1)
                self.geom = Cline(cellcard.str[:start])
                self.parm = Cline(cellcard.str[start:])

                # look for transformation in cell parameters
                self.parm.remove_comments()
                m = trans.search(self.parm.str)
                if m:
                    self.hproc = False
                self.parm.restore_comments()
            else:
                self.geom = Cline(cellcard.str)
                self.parm = Cline("")

        return

    def __get_data__(self):

        mt = celtrsf.search(self.parm.str)
        if mt:
            string = self.parm.str[mt.end() :]
            m = grlnumber.search(string)

            if self.parm.str[mt.start() - 1] == "*":
                angle = True
            else:
                angle = False

            if "(" in string[: m.end()]:
                self.TRCL = []
                pos = 0
                while True:
                    m = trfnumber.search(string, pos)
                    pos = m.end()
                    if not m or m.group() == ")":
                        break
                    self.TRCL.append(float(m.group()))
                if angle and len(self.TRCL) > 3:
                    self.TRCL[3:12] = list(map(math.radians, self.TRCL[3:12]))
                    self.TRCL[3:12] = list(map(math.cos, self.TRCL[3:12]))
            else:
                self.TRCL = int(m.group())

        # get universe number
        m = celuniverse.search(self.parm.str)
        if m:
            m = grlnumber.search(self.parm.str[m.end() :])
            self.U = int(m.group())

        # get Fill number
        mf = celfill.search(self.parm.str)
        if mf:
            fillstring = remove_dollar(self.parm.str[mf.end() :])
            m = grlnumber.search(fillstring)
            self.FILL = int(m.group())

            # look for transformation after fill card
            m = re.search(r" *\(", fillstring)
            if m:
                if self.parm.str[mf.start() - 1] == "*":
                    angle = True
                else:
                    angle = False

                fillstring = fillstring[m.end() :]
                m = re.search(r"\)", fillstring)
                string = fillstring[: m.start()]
                tr = string.split()
                if len(tr) == 1:
                    self.TR = int(tr[0])
                else:
                    self.TR = list(map(float, tr))
                    if angle and len(self.TR) > 3:
                        self.TR[3:12] = list(map(math.radians, self.TR[3:12]))
                        self.TR[3:12] = list(map(math.cos, self.TR[3:12]))

    def get_stat(self, remove_com=True):
        """Count and return the number of words and hashes on the line."""

        if remove_com:
            self.geom.remove_comments()

        words = len(reword.findall(self.geom.str))
        hashcell = len(compcell.findall(self.geom.str))
        hashtot = self.geom.str.count("#")

        self.stat = {
            "words": words,
            "hash": hashtot,
            "hascell": hashcell,
            "hashsur": hashtot - hashcell,
        }
        if remove_com:
            self.geom.restore_comments()
        return [words, hashtot, hashcell, hashtot - hashcell]

    def get_lines(self):
        """split string card in the format Cards.lines of Cards object"""
        card = self.headstr + self.geom.str + self.parm.str

        # remove blank line introduced during the process
        card = re.sub(blnkline, "", card)

        # subtitute comment $ with  blank line
        card = re.sub(comdollar, r"\nC\g<blnk>", card)

        # ensure 5 blanks continous line
        card = re.sub(contline, r"\n     \g<start>", card)

        if card[-1] == "\n":
            card = card[:-1]
        return list(map(lambda x: x + "\n", card.split("\n")))


def remove_hash(cards, cname, keepComments=True):
    def remove(card, cname, keepComments):
        """remove complementary operator and subtitute by complementary cell"""
        if "parser.Card" in str(type(card)):
            celline = "".join(card.lines)
            cardstr = CellCardString(celline)
        else:
            cardstr = card
        cardstr.get_stat()
        cardstr.geom.remove_comments(full=not keepComments)
        if (not cardstr.hproc) or (cardstr.stat["hash"] == 0):
            return cardstr.geom  # no complementary operator or cannot be # cannot be removed
        cell = Cline(cardstr.geom.str)
        # find all operators in the cell and
        # substitute all complementary operators
        # locate object in list to reverse iteration
        hashgroup = []
        start = 0

        lencel = len(cell.str)
        while True:
            ic = cell.str.lower().find("c", start)
            idol = cell.str.find("$", start)
            if idol < 0:
                idol = lencel
            if ic < 0:
                ic = lencel
            end = min(idol, ic)
            for m in rehash.finditer(cell.str, start, end):
                hashgroup.append(m)
            start = cell.str.find("\n", end)
            if end == lencel:
                break

        for m in reversed(hashgroup):
            start = m.start()
            if m.group(1) == "(":  # complementary cell defined as surface intersections
                hcell, end = cell.get_hashcell(start)
                cellmod = cell.str[0:start] + complementary(hcell) + cell.str[end:]
                cell = Cline(cellmod)
            else:
                hcname = int(m.group(1))  # complementary cell defined with other cell index
                newdef = remove(cards[hcname], hcname, keepComments)  # remove complementary operator in new cell if necessary
                end = m.end()
                cellmod = cell.str[0:start] + "      " + complementary(newdef) + " " + cell.str[end:]
                cell = Cline(cellmod)
        return cell

    newcell = remove(cards[cname], cname, keepComments)
    return Cline(newcell.str)



#
# Module to load a STEP file
#
import logging
import os
import re

import FreeCAD
import Part
from FreeCAD import Import

from ..utils import functions as UF
from . import load_functions as LF

logger = logging.getLogger("general_logger")


# Paco mod
def extract_materials(filename):
    rho_real = []
    m_dict = {}  # _ Material dictionary
    with open(filename, "rt") as file:
        for line in file:
            vals = line.split()
            if vals[0].startswith("#"):
                continue
            mat_label = int(vals[0])
            rho_real = -float(vals[1])
            matname = " ".join(vals[2:])
            m_dict[mat_label] = (rho_real, matname)
    return m_dict


def load_cad(filename, settings, options):

    # Set document solid tree options when opening CAD differing from version 0.18
    if int(FreeCAD.Version()[1]) > 18:
        LF.set_doc_options()

    cad_simplificado_doc = FreeCAD.newDocument("CAD_simplificado")
    Import.insert(filename, "CAD_simplificado")

    if settings.matFile != "":
        if os.path.exists(settings.matFile):
            m_dict = extract_materials(settings.matFile)
        else:
            logger.info(f"Material definition file {settings.matFile} does not exist.")
            m_dict = {}
    else:
        m_dict = {}

    s = Part.Shape()
    s.read(filename)
    Solids = s.Solids
    meta_list = []
    for i, s in enumerate(Solids):
        meta_list.append(UF.GeounedSolid(i + 1, s))

    i_solid = 0
    missing_mat = set()

    doc_objects = cad_simplificado_doc.Objects

    for elem in doc_objects:
        if elem.TypeId == "Part::Feature":
            comment = LF.getCommentTree(elem, options)
            if not elem.Shape.Solids:
                logger.warning("Element {:} has no associated solid".format(comment + "/" + elem.Label))
                continue
            else:
                tempre_mat = None
                tempre_dil = None

                # MIO: lightly modification of label if required
                label = LF.get_label(elem.Label, options)
                comment = comment + "/" + label
                if elem.InList:
                    # MIO: lightly modification of label if required
                    label_in_list = LF.get_label(elem.InList[0].Label, options)
                    encl_label = re.search("enclosure(?P<encl>[0-9]+)_(?P<parent>[0-9]+)_", label_in_list)
                    if not encl_label:
                        encl_label = re.search("enclosure(?P<encl>[0-9]+)_(?P<parent>[0-9]+)_", label)

                    envel_label = re.search("envelope(?P<env>[0-9]+)_(?P<parent>[0-9]+)_", label_in_list)
                    if not envel_label:
                        envel_label = re.search("envelope(?P<env>[0-9]+)_(?P<parent>[0-9]+)_", label)

                    # tempre_mat = re.search("(m(?P<mat>\d+)_)",elem.Label)
                    # if not tempre_mat :
                    #    tempre_mat = re.search("(m(?P<mat>\d+)_)",elem.InList[0].Label)

                    # tempre_dil = re.search("(_d(?P<dil>\d*\.\d*)_)",elem.Label)
                    # if not tempre_dil :
                    #    tempre_dil = re.search("(_d(?P<dil>\d*\.\d*)_)",elem.InList[0].Label)

                    # Paco modifications
                    # Search for material definition in tree
                    xelem = [elem]
                    while xelem and not tempre_mat:
                        # MIO: Modification of label if required
                        temp_label = LF.get_label(xelem[0].Label, options)
                        tempre_mat = re.search("_m(?P<mat>\d+)_", "_" + temp_label)
                        xelem = xelem[0].InList

                    # Search for dilution definition in tree
                    xelem = [elem]
                    while xelem and not tempre_dil:
                        # MIO: Modification of label if required
                        temp_label = LF.get_label(xelem[0].Label, options)
                        tempre_dil = re.search("_d(?P<dil>\d*\.\d*)_", temp_label)
                        xelem = xelem[0].InList
                    # Paco end
                else:
                    encl_label = None
                    envel_label = None

                # compSolid Diferent solid of the same cell are stored in the same metaObject (compSolid)
                # enclosures and envelopes are always stored as compound
                if settings.compSolids or encl_label or envel_label:

                    init = i_solid
                    end = i_solid + len(elem.Shape.Solids)
                    LF.fuse_meta_obj(meta_list, init, end)
                    n_solids = 1
                else:
                    n_solids = len(elem.Shape.Solids)

                for i in range(n_solids):
                    meta_list[i_solid].set_comments(f"{comment}{i + 1}")
                    meta_list[i_solid].set_cad_solid()

                    if tempre_mat:
                        mat_label = int(tempre_mat.group("mat"))
                        if mat_label in m_dict.keys():
                            meta_list[i_solid].set_material(mat_label, m_dict[mat_label][0], m_dict[mat_label][1])
                        else:
                            if mat_label == 0:
                                meta_list[i_solid].set_material(mat_label, 0, 0)
                            else:
                                meta_list[i_solid].set_material(
                                    mat_label,
                                    -100,
                                    "Missing material density information",
                                )
                                missing_mat.add(mat_label)
                    else:
                        # logger.warning('No material label associated to solid {}.\nDefault material used instead.'.format(comment))
                        if settings.voidMat:
                            meta_list[i_solid].set_material(*settings.voidMat)
                    if tempre_dil:
                        meta_list[i_solid].set_dilution(float(tempre_dil.group("dil")))

                    if encl_label is not None:
                        meta_list[i_solid].EnclosureID = int(encl_label.group("encl"))
                        meta_list[i_solid].ParentEnclosureID = int(encl_label.group("parent"))
                        meta_list[i_solid].IsEnclosure = True
                        meta_list[i_solid].CellType = "void"

                    if envel_label is not None:
                        meta_list[i_solid].EnclosureID = int(envel_label.group("env"))
                        meta_list[i_solid].ParentEnclosureID = int(envel_label.group("parent"))
                        meta_list[i_solid].IsEnclosure = True
                        meta_list[i_solid].CellType = "envelope"
                    i_solid += 1

    LF.joinEnvelopes(meta_list)
    if missing_mat:
        logger.warning("At least one material in the CAD model is not present in the material file")
        logger.info(f"List of not present materials: {missing_mat}")

    enclosure_list = LF.set_enclosure_solid_list(meta_list)
    if enclosure_list:
        LF.check_enclosure(cad_simplificado_doc, enclosure_list)
        # LF.remove_enclosure(meta_list)
        return meta_list, enclosure_list
    else:
        return meta_list, []



import math

import BOPTools.SplitAPI
import FreeCAD
import Part


class SplitBase:
    def __init__(self, base, knownSurf={}):
        self.base = base
        self.knownSurf = knownSurf


def joinBase(baseList):
    shape = []
    surf = {}
    removedKeys = []
    for b in baseList:
        if b.base is not None:
            shape.append(b.base)
        for k, v in b.knownSurf.items():
            if k in removedKeys:
                continue
            if k not in surf.keys():
                surf[k] = v
            else:
                if surf[k] == v:
                    continue
                else:
                    surf[k] = None
                    removedKeys.append(k)

    newbase = FuseSolid(shape)
    return SplitBase(newbase, surf)


# TODO rename this function as there are two with the name name
def SplitSolid(base, surfacesCut, cellObj, solidTool=False, tolerance=0.01):  # 1e-2
    # split Base (shape Object or list/tuple of shapes)
    # with selected surfaces (list of surfaces objects) cutting the base(s) (surfacesCut)
    # cellObj is the CAD object of the working cell to reconstruction.
    # the function return a list of solids enclosed fully in the cell (fullPart)
    # and a list of solids not fully enclosed in the cell (cutPart). These lasts
    # will require more splitting with the others surfaces defining the cell.

    fullPart = []
    cutPart = []

    # part if several base in input
    if type(base) is list or type(base) is tuple:
        for b in base:
            fullList, cutList = SplitSolid(b, surfacesCut, cellObj, tolerance=tolerance)
            fullPart.extend(fullList)
            cutPart.extend(cutList)
        return fullPart, cutPart

    # part if base is shape object

    if solidTool:
        Tools = (cellObj.shape,)
    else:
        Tools = tuple(s.shape for s in surfacesCut)
    # for s in surfacesCut:
    #    print(s.type,s.params,s.id)
    #    s.shape.exportStep('tool{}.stp'.format(s.id))
    # base.base.exportStep('base.stp')
    Solids = BOPTools.SplitAPI.slice(base.base, Tools, "Split", tolerance=tolerance).Solids
    if not Solids:
        Solids = [base.base]
    partPositions, partSolids = space_decomposition(Solids, surfacesCut)

    for pos, sol in zip(partPositions, partSolids):
        # fullPos = updateSurfacesValues(pos,cellObj.surfaces,base.knownSurf)
        # inSolid = cellObj.definition.evaluate(fullPos)

        if not solidTool:
            pos.update(base.knownSurf)
        inSolid = cellObj.definition.evaluate(pos)

        # if solidTool :
        #  ii += 1
        #  print(solidTool)
        #  print(cellObj.definition)
        #  print(pos)
        #  print('eval',inSolid)
        #  name = str(cellObj.definition)
        #  sol.exportStep('solid_{}{}.stp'.format(name,ii))

        if inSolid:
            fullPart.append(SplitBase(sol, pos))
        elif inSolid is None:
            cutPart.append(SplitBase(sol, pos))
    return fullPart, cutPart


def updateSurfacesValues(position, surfaces, knownSurf):
    position.update(knownSurf)
    sname = set(surfaces.keys())
    pname = set(position.keys())

    fullpos = position.copy()
    for name in sname.difference(pname):
        fullpos[name] = None
    return fullpos


# Get the position of subregion with respect
# all cutting surfaces
def space_decomposition(solids, surfaces):

    component = []
    good_solids = []
    for c in solids:
        if c.Volume < 1e-3:
            if abs(c.Volume) < 1e-3:
                continue
            else:
                c.reverse()
                print("Negative solid Volume", c.Volume)
        Svalues = {}
        point = point_inside(c)
        if point == None:
            continue  # point not found in solid (solid is surface or very thin can be source of lost particules in MCNP)
        for surf in surfaces:
            Svalues[surf.id] = surface_side(point, surf)

        component.append(Svalues)
        good_solids.append(c)
    return component, good_solids


# find one point inside a solid (region)
def point_inside(solid):

    cut_line = 32
    cut_box = 4

    # no poner boundbox, el punto puente caer en una superficie para geometria triangular
    point = solid.CenterOfMass
    if solid.isInside(point, 0.0, False):
        return point

    v1 = solid.Vertexes[0].Point
    for vi in range(len(solid.Vertexes) - 1, 0, -1):
        v2 = solid.Vertexes[vi].Point
        dv = (v2 - v1) * 0.5

        n = 1
        while True:
            for i in range(n):
                point = v1 + dv * (1 + 0.5 * i)
                if solid.isInside(point, 0.0, False):
                    return point
            n = n * 2
            dv = dv * 0.5
            if n > cut_line:
                break

    #      Box_Volume = BBox.XLength*BBox.YLength*BBox.ZLength
    #      if (solid.Volume < Box_Volume/ math.pow(16,nmax_cut)) :
    #           print('very small Solid Volume (solid volume, box volume): {},{}'.format(solid.Volume,Box_Volume))
    #           return None
    BBox = solid.optimalBoundingBox(False)
    box = [BBox.XMin, BBox.XMax, BBox.YMin, BBox.YMax, BBox.ZMin, BBox.ZMax]

    boxes, centers = divide_box(box)
    n = 0

    while True:
        for p in centers:
            pp = FreeCAD.Vector(p[0], p[1], p[2])
            if solid.isInside(pp, 0.0, False):
                return pp

        subbox = []
        centers = []
        for b in boxes:
            btab, ctab = divide_box(b)
            subbox.extend(btab)
            centers.extend(ctab)
        boxes = subbox
        n = n + 1

        if n == cut_box:
            print(f"Solid not found in bounding Box (Volume : {solid.Volume})")
            print("Valid Solid : ", solid.isValid())
            return None


# divide a box into 8 smaller boxes
def divide_box(Box):
    xmid = (Box[1] + Box[0]) * 0.5
    ymid = (Box[3] + Box[2]) * 0.5
    zmid = (Box[5] + Box[4]) * 0.5

    b1 = (Box[0], xmid, Box[2], ymid, Box[4], zmid)
    p1 = (0.5 * (Box[0] + xmid), 0.5 * (Box[2] + ymid), 0.5 * (Box[4] + zmid))

    b2 = (xmid, Box[1], Box[2], ymid, Box[4], zmid)
    p2 = (0.5 * (xmid + Box[1]), 0.5 * (Box[2] + ymid), 0.5 * (Box[4] + zmid))

    b3 = (Box[0], xmid, ymid, Box[3], Box[4], zmid)
    p3 = (0.5 * (Box[0] + xmid), 0.5 * (ymid + Box[3]), 0.5 * (Box[4] + zmid))

    b4 = (xmid, Box[1], ymid, Box[3], Box[4], zmid)
    p4 = (0.5 * (xmid + Box[1]), 0.5 * (ymid + Box[3]), 0.5 * (Box[4] + zmid))

    b5 = (Box[0], xmid, Box[2], ymid, zmid, Box[5])
    p5 = (0.5 * (Box[0] + xmid), 0.5 * (Box[2] + ymid), 0.5 * (zmid + Box[5]))

    b6 = (xmid, Box[1], Box[2], ymid, zmid, Box[5])
    p6 = (0.5 * (xmid + Box[1]), 0.5 * (Box[2] + ymid), 0.5 * (zmid + Box[5]))

    b7 = (Box[0], xmid, ymid, Box[3], zmid, Box[5])
    p7 = (0.5 * (Box[0] + xmid), 0.5 * (ymid + Box[3]), 0.5 * (zmid + Box[5]))

    b8 = (xmid, Box[1], ymid, Box[3], zmid, Box[5])
    p8 = (0.5 * (xmid + Box[1]), 0.5 * (ymid + Box[3]), 0.5 * (zmid + Box[5]))

    return [b1, b2, b3, b4, b5, b6, b7, b8], [p1, p2, p3, p4, p5, p6, p7, p8]


# check the position of the point with respect
# a surface
def surface_side(p, surf):
    if surf.type == "sphere":
        org, R = surf.params
        D = p - org
        inout = D.Length - R

    elif surf.type == "plane":
        normal, d = surf.params
        inout = p.dot(normal) - d

    elif surf.type == "cylinder":
        P, v, R = surf.params

        D = p - P
        if not surf.truncated:
            inout = D.cross(v).Length - R
        else:
            inCyl = D.cross(v).Length / v.Length - R  # <0 in cylinder
            inPln = btwPPlanes(p, P, v)  # <0  between planes

            if (inCyl < 0) and (inPln < 0):
                inout = -1  # inside the can
            else:
                inout = 1  # outside the can

    elif surf.type == "cone":
        if not surf.truncated:
            P, v, t, dblsht = surf.params
            X = p - P
            X.normalize()
            dprod = X.dot(v)
            dprod = max(-1, min(1, dprod))
            a = math.acos(dprod) if not dblsht else math.acos(abs(dprod))
            inout = a - math.atan(t)
        else:
            P, v, R1, R2 = surf.params
            apex = P + R1 / (R1 - R2) * v

            X = p - apex
            X.normalize()
            dprod = X.dot(-v) / v.Length  # -v because reverse axis. in MCNP TRC r1 > r2
            dprod = max(-1, min(1, dprod))
            a = math.acos(dprod)

            t = (R1 - R2) / v.Length
            inCone = a - math.atan(t)
            inPln = btwPPlanes(p, P, v)  # <0  between planes

            if (inCone < 0) and (inPln < 0):
                inout = -1  # inside the can
            else:
                inout = 1  # outside the can

    elif surf.type == "cone_elliptic":
        apex, axis, Ra, radii, rAxes, dblsht = surf.params

        r = p - apex
        X = r.dot(rAxes[1])
        Y = r.dot(rAxes[0])
        Z = r.dot(axis)
        if dblsht:
            Z = abs(Z)
        inout = (X / radii[1]) ** 2 + (Y / radii[0]) ** 2 - Z / Ra

    elif surf.type == "hyperboloid":
        center, axis, radii, rAxes, onesht = surf.params

        r = p - center
        rX = r.dot(rAxes[1])
        v = r - (rX * rAxes[1] + center)
        d = v.Length

        one = 1 if onesht else -1
        radical = (rX / radii[1]) ** 2 + one

        if radical > 0:
            Y = radii[0] * math.sqrt(radical)
            inout = d - Y
        else:
            inout = 1

    elif surf.type == "ellipsoid":
        center, axis, radii, rAxes = surf.params

        r = p - center
        rX = r.dot(axis)
        rY = r - (rX * axis + center)

        if axis.add(-rAxes[0]).Length < 1e-5:
            radX, radY = radii
        else:
            radY, radY = radii

        radical = 1 - (rX / radX) ** 2
        if radical > 0:
            Y = radY * math.sqrt(radical)
            inout = rY - Y
        else:
            inout = 1

    elif surf.type == "cylinder_elliptic":
        center, axis, radii, rAxes = surf.params

        r = p - center
        X = r.dot(rAxes[1])
        Y = r.dot(rAxes[0])
        inout = (X / radii[1]) ** 2 + (Y / radii[0]) ** 2 - 1

        if surf.truncated and inout < 0:
            inout = btwPPlanes(p, center, axis)  # <0  between planes

    elif surf.type == "cylinder_hyperbolic":
        center, axis, radii, rAxes = surf.params

        r = p - center
        X = r.dot(rAxes[1])
        Y = r.dot(rAxes[0])
        inout = (X / radii[1]) ** 2 - (Y / radii[0]) ** 2 - 1

    elif surf.type == "paraboloid":
        center, axis, focal = surf.params

        r = p - center
        X = r.dot(axis)
        if X < 0:
            inout = 1
        else:
            v = r - X * axis
            d = v.Length
            Y = math.sqrt(4 * focal * X)
            inout = d - Y

    elif surf.type == "torus":
        P, v, Ra, Rb, Rc = surf.params

        d = p - P
        z = d.dot(v)
        rz = d - z * v
        inout = (z / Rb) ** 2 + ((rz.Length - Ra) / Rc) ** 2 - 1

    elif surf.type == "box":
        P, v1, v2, v3 = surf.params
        for v in (v1, v2, v3):
            inout = btwPPlanes(p, P, v)  # <0  between planes
            if inout > 0:
                break

    else:
        print(f"surface type {surf[0]} not considered")
        return

    return inout > 0


def btwPPlanes(p, p0, v):

    p1 = p0 + v
    inP0 = v.dot(p - p0)  # >0 plane(base plane) side inside the cylinder
    inP1 = v.dot(p - p1)  # >0 plane(base plane) side inside the cylinder

    if (inP0 > 0) and (inP1 < 0):
        return -1
    else:
        return 1


# ************************************************


def FuseSolid(parts):
    if (len(parts)) <= 1:
        if parts:
            solid = parts[0]
        else:
            return None
    else:
        try:
            fused = parts[0].fuse(parts[1:])
        except:
            fused = None

        if fused is not None:
            try:
                refinedfused = fused.removeSplitter()
            except:
                refinedfused = fused

            if refinedfused.isValid():
                solid = refinedfused
            else:
                if fused.isValid():
                    solid = fused
                else:
                    solid = Part.makeCompound(parts)
        else:
            solid = Part.makeCompound(parts)

    if solid.Volume < 0:
        solid.reverse()
    return solid



from pathlib import Path


def comments_write(name, MetaList):
    """Function to write in an independent file the comment strings"""

    Path(name).parent.mkdir(parents=True, exist_ok=True)
    with open(file=name + "_comments.txt", mode="w", encoding="utf-8") as outfile:
        for m in MetaList:
            outfile.write(m.Comments + "\n")


def summary_write(name, MetaList):

    Path(name).parent.mkdir(parents=True, exist_ok=True)
    with open(file=name + "_summary.txt", mode="w", encoding="utf-8") as outfile:
        header = f"  Cell Id{'':5s}Mat Id{'':6s}Density{'':7s}Volume{'':5s}Comments\n"
        outfile.write(header)

        for m in MetaList:
            if m.Void or m.__id__ is None:
                continue
            index = m.__id__
            Vol = m.Volume * 1e-3
            if m.Material == 0:
                line = " {:>8d}{:3s}{:>8d}{:3s}{:11s}{:3s}{:11.4e}{:3s}{}\n".format(
                    index, "", 0, "", "", "", Vol, "", m.Comments
                )
            else:
                if abs(m.Density) < 1e-2:
                    line = " {:>8d}{:3s}{:>8d}{:3s}{:11.4e}{:3s}{:11.4e}{:3s}{}\n".format(
                        index,
                        "",
                        m.Material,
                        "",
                        m.Density,
                        "",
                        Vol,
                        "",
                        m.Comments,
                    )
                else:
                    line = " {:>8d}{:3s}{:>8d}{:3s}{:11.7f}{:3s}{:11.4e}{:3s}{}\n".format(
                        index,
                        "",
                        m.Material,
                        "",
                        m.Density,
                        "",
                        Vol,
                        "",
                        m.Comments,
                    )

            outfile.write(line)



#!/usr/bin/env python3

"""
Convert Cad geometry to CSG format for use in Monte Carlo Codes
"""

import argparse
import json

from pathlib import Path

import geouned

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-i",
    "--input",
    type=str,
    default="config.json",
    help="The path to the config JSON file",
)
args = parser.parse_args()


def main():

    if not Path(args.input).exists():
        raise FileNotFoundError(f"config file {args.input} not found")

    with open(args.input) as f:
        config = json.load(f)

    geo = geouned.CsgToCad()
    geo.export_cad(**config["export_cad"])


if __name__ == "__main__":
    main()



from .remh import Cline


class CellCard:

    def __init__(self, data):
        self.type = "cell"
        self.TR = None
        self.processData(data)

    def processData(self, data):
        self.name = int(data["id"])
        self.level = None

        if "material" in data.keys():
            self.MAT = 0 if data["material"] == "void" else int(data["material"])
        else:
            self.MAT = None

        if "universe" in data.keys():
            self.U = 0 if int(data["universe"]) == 1 else int(data["universe"])
        else:
            self.U = 0

        if "fill" in data.keys():
            self.FILL = int(data["fill"])
        else:
            self.FILL = None

        self.geom = Cline(data["region"].replace("|", ":"))


class SurfCard:
    def __init__(self, data):

        self.type = "surface"
        self.processData(data)

    def processData(self, data):
        self.name = int(data["id"])
        self.stype = data["type"]
        self.scoefs = tuple(float(x) for x in data["coeffs"].split())


def get_cards(root):
    for c in root:
        yield process_card(c)


def process_card(card):
    ctype = card.tag
    if ctype == "cell":
        return CellCard(card.attrib)

    elif ctype == "surface":
        return SurfCard(card.attrib)



#
# Define the class storing Boolean expression of the cell definition
#

import logging
import re

logger = logging.getLogger("general_logger")

mostinner = re.compile(r"\([^\(^\)]*\)")  # identify most inner parentheses
mix = re.compile(r"(?P<value>([-+]?\d+|\[0+\]))")  # identify signed integer or [000...] pattern. Record the value.
TFX = re.compile(r"(?P<value>[FTXo]+)")  # identify pattern including F,T,X, or o sequence ( in any order).


class BoolSequence:
    """Class storing Boolean expression and operating on it"""

    def __init__(self, definition=None, operator=None):
        if definition:
            self.elements = []
            self.set_def(definition)
        else:
            self.elements = []
            self.operator = operator
            self.level = 0

    def __str__(self):
        out = f"{self.operator}["
        if type(self.elements) is bool:
            return " True " if self.elements else " False "
        for e in self.elements:
            if type(e) is int or type(e) is bool or type(e) is str:
                out += f" {e} "
            else:
                out += e.__str__()

        out += "] "
        return out

    def append(self, *seq):
        """Append a BoolSequence Objects. seq may be :
        - An iterable containing allowed BoolSequence Objects
        - A BoolSequence object
        - An integer value
        - A Boolean value"""

        if type(self.elements) is bool:
            if (self.elements and self.operator == "AND") or (not self.elements and self.operator == "OR"):
                self.assign(seq)
            return

        for s in seq:
            if type(s) is int:
                level = -1
                if s in self.elements:
                    continue
                elif -s in self.elements:
                    self.level = -1
                    if self.operator == "AND":
                        self.elements = False
                    else:
                        self.elements = True
                    return
            elif type(s) is bool:
                if self.operator == "AND" and s or self.operator == "OR" and not s:
                    continue
                else:
                    self.elements = s
                    self.level = -1
                    return
            else:
                level = s.level
                if type(s.elements) is bool:
                    if self.operator == "AND" and not s.elements or self.operator == "OR" and s.elements:
                        self.level = -1
                        self.elements = s.elements
                        return
                    else:
                        continue

            self.elements.append(s)
            self.level = max(self.level, level + 1)

    def assign(self, seq):
        """Assign the BoolSequence Seq to the self instance BoolSequence"""
        if type(seq) is bool:
            self.operator == "AND"
            self.elements = seq
            self.level = -1
            return

        self.operator = seq.operator
        self.elements = seq.elements
        self.level = seq.level

    def update(self, seq, pos):
        if len(pos) == 0:
            self.assign(seq)
            return
        elif len(pos) == 1:
            base = self
        else:
            base = self.get_element(pos[:-1])

        indexes = pos[-1]
        indexes.sort()
        for i in reversed(indexes):
            del base.elements[i]

        if type(seq.elements) is bool:
            base.elements = seq.elements
            base.level = -1
        else:
            base.append(seq)
            base.join_operators()
        self.clean(self_level=True)
        return

    def get_element(self, pos):
        if len(pos) == 1:
            return self.elements[pos[0]]
        else:
            return self.elements[pos[0]].get_element(pos[1:])

    def copy(self):
        cp = BoolSequence()
        cp.operator = self.operator
        cp.level = self.level
        if type(self.elements) is bool:
            cp.elements = self.elements
        else:
            for e in self.elements:
                if type(e) is int:
                    cp.elements.append(e)
                else:
                    cp.elements.append(e.copy())
        return cp

    # TODO rename to snake case, care as multiple functions with same name
    def get_complementary(self):
        c = BoolSequence(operator=self.comp_operator())
        c.level = self.level

        if self.level == 0:
            for e in self.elements:
                c.elements.append(-e)
            return c
        else:
            self.group_single()
            for e in self.elements:
                c.elements.append(e.get_complementary())
            return c

    def comp_operator(self):
        if self.operator == "AND":
            return "OR"
        else:
            return "AND"

    def simplify(self, CT=None, depth=0):
        """Simplification by recursive calls to the inner BoolSequence objects."""
        if self.level > 0:
            for seq in self.elements:
                seq.simplify(CT, depth + 1)
            self.clean()
            self.join_operators()
            self.level_update()

        if type(self.elements) is not bool and (self.level > 0 or len(self.elements) > 1):
            levIn = self.level
            self.simplify_sequence(CT)

            if self.level > levIn and depth < 10:
                self.simplify(CT, depth + 1)

    def simplify_sequence(self, CT=None):
        """Carry out the simplification process of the BoolSequence."""
        if self.level < 1 and CT is None:
            self.clean()
            return

        surf_names = self.get_surfaces_numbers()
        if not surf_names:
            return

        newNames = surf_names
        for val_name in surf_names:
            if val_name in newNames:

                if CT is None:
                    true_set = {abs(val_name): True}
                    false_set = {abs(val_name): False}
                else:
                    true_set, false_set = CT.get_constraint_set(val_name)

                if not self.do_factorize(val_name, true_set, false_set):
                    continue
                self.factorize(val_name, true_set, false_set)
                if type(self.elements) is bool:
                    return
                newNames = self.get_surfaces_numbers()

    def do_factorize(self, val_name, true_set, false_set):
        """For level 0 sequence check if the factorization would lead to a simplification."""
        if self.level > 0:
            return True
        if true_set is None and false_set is None:
            logger.info(f"{val_name} is not true nor false")
            return False
        if true_set is None or false_set is None:
            return True

        val_set = self.get_surfaces_numbers()
        t_set = set(true_set.keys()) & val_set
        f_set = set(false_set.keys()) & val_set

        if len(t_set) == 1 and len(f_set) == 1:
            return False

        value = None
        for val in self.elements:
            if abs(val) == val_name:
                value = val
                break

        if value is None:
            return False

        if len(t_set) == 1:
            if self.operator == "AND":
                # if value > 0 and t_set[val_name] or value < 0 and not t_set[val_name] : return False
                if value > 0:
                    return False  # TrueSet[Valname] always True
            else:
                # if value < 0 and t_set[val_name] or value > 0 and not t_set[val_name] : return False
                if value < 0:
                    return False

        elif len(f_set) == 1:
            if self.operator == "AND":
                # if value > 0 and f_set[val_name] or value < 0 and not f_set[val_name] : return False
                if value < 0:
                    return False
            else:
                # if value < 0 and f_set[val_name] or value > 0 and not f_set[val_name] : return False
                if value > 0:
                    return False

        return True

    # check if level 0 sequence have opposite value a & -a = 0  , a|-a = 1
    # return the value of the sequence None(unknown), True, False
    def check(self, level0=False):
        """Check BoolSequence in level 0 have oposite values  a & -a = 0  , a|-a = 1."""
        if type(self.elements) is bool:
            return self.elements
        if self.level == 0:
            signed_surf = set(self.elements)
            surf_name = self.get_surfaces_numbers()
            if len(signed_surf) == len(surf_name):
                return None  # means same surface has not positive and negative value
            elif self.operator == "AND":
                self.elements = False
                self.level = -1
                return False
            else:
                self.elements = True
                self.level = -1
                return True
        elif not level0:
            self.group_single()
            none_val = False
            for e in reversed(self.elements):
                e.check()
                if type(e.elements) is bool:
                    res = e.elements
                else:
                    res = None

                if res is None:
                    none_val = True
                elif self.operator == "AND" and res is False:
                    self.level = -1
                    self.elements = False
                    return False
                elif self.operator == "OR" and res is True:
                    self.level = -1
                    self.elements = True
                    return True
                else:
                    self.elements.remove(e)

            if none_val:
                return None
            elif self.operator == "AND":
                self.level = -1
                self.elements = True
                return True
            else:
                self.level = -1
                self.elements = False
                return False

    def substitute(self, var, val):
        """Substitute in the BoolSequence the variable "var" by the value "val".
        "val" can be an Boolean value or an integer representing another surface variable.
        """
        if val is None:
            return
        if type(self.elements) is bool:
            return
        name = abs(var)
        ic = len(self.elements)
        for e in reversed(self.elements):
            ic -= 1
            if type(e) is int:
                if abs(e) == name:
                    if type(val) is int:
                        if name == e:
                            self.elements[ic] = val
                        else:
                            self.elements[ic] = -val

                    else:
                        if name == e:
                            bool_value = val
                        else:
                            bool_value = not val

                        if self.operator == "AND" and not bool_value:
                            self.elements = False
                            self.level = -1
                            return
                        elif self.operator == "OR" and bool_value:
                            self.elements = True
                            self.level = -1
                            return
                        else:
                            self.elements.remove(e)

            else:
                e.substitute(var, val)

        if self.elements == []:
            self.elements = True if self.operator == "AND" else False
            self.level = -1
            return

        self.clean(self_level=True)
        self.check(level0=True)
        self.join_operators(self_level=True)

    def clean(self, self_level=False):
        """Remove sequences whom elements are boolean values instead of list."""
        if type(self.elements) is bool:
            return self.elements
        for e in reversed(self.elements):
            if type(e) is int:
                continue
            eVal = e if self_level else e.clean()
            if type(eVal) is not bool:
                eVal = eVal.elements

            if type(eVal) is bool:
                if eVal and self.operator == "OR":
                    self.elements = True
                    self.level = -1
                    return True
                elif not eVal and self.operator == "AND":
                    self.elements = False
                    self.level = -1
                    return False
                self.elements.remove(e)

        if self.elements == []:
            if self.operator == "OR":
                self.elements = False
            else:
                self.elements = True
            self.level = -1
            return self.elements
        else:
            return self

    # TODO rename to snake case, care as multiple functions with same name
    def join_operators(self, self_level=False):
        """Join redundant operators in found in the sequence."""
        if type(self.elements) is bool:
            return
        self.clean(self_level=True)
        self.level_update()
        if self.level == 0:
            return
        self.group_single()
        ANDop = []
        ORop = []

        for e in self.elements:
            if e.operator == "AND":
                ANDop.append(e)
            else:
                ORop.append(e)

        if len(ANDop) > 1 and self.operator == "AND":
            newSeq = BoolSequence(operator="AND")
            for s in ANDop:
                newSeq.elements.extend(s.elements)
                self.elements.remove(s)
            newSeq.level_update()
            self.append(newSeq)

        elif len(ORop) > 1 and self.operator == "OR":
            newSeq = BoolSequence(operator="OR")
            for s in ORop:
                newSeq.elements.extend(s.elements)
                self.elements.remove(s)
            newSeq.level_update()
            self.append(newSeq)

        if self.level > 0 and len(self.elements) == 1:
            self.operator = self.elements[0].operator
            self.elements[:] = self.elements[0].elements[:]
            self.level -= 1
            self.join_operators()

        if self.level == 0:
            self.check()
            return

        if not self_level:
            if type(self.elements) is bool:
                return
            for e in self.elements:
                e.join_operators()

    def get_sub_sequence(self, setIn):
        if type(setIn) is set:
            val_set = setIn
        elif type(setIn) is int:
            val_set = {setIn}
        else:
            val_set = set(setIn.keys())

        if self.level == 0:
            return ([], self)

        position = []
        subSeq = BoolSequence(operator=self.operator)

        for pos, e in enumerate(self.elements):
            surf = e.get_surfaces_numbers()
            if len(surf & val_set) != 0:
                subSeq.append(e)
                position.append(pos)

        if len(position) == 1 and subSeq.elements[0].level > 0:
            subList, subSeq = subSeq.elements[0].get_sub_sequence(val_set)
            subList.insert(0, position[0])
        else:
            subList = [position]

        return subList, subSeq

    def factorize(self, valname, true_set, false_set):
        """Make the factorization of the Sequence on variable valname using Shannon's theorem."""
        if true_set is None:  # valname cannot take True value
            falseFunc = self.evaluate(false_set)
            self.assign(falseFunc)
            return True

        if false_set is None:  # valname cannot take false value
            trueFunc = self.evaluate(true_set)
            self.assign(trueFunc)
            return True

        val_set = set(true_set.keys())
        val_set.update(false_set.keys())
        pos, subSeq = self.get_sub_sequence(val_set)
        updt = True
        if len(pos) == 0:
            subSeq = self
            updt = False

        trueFunc = subSeq.evaluate(true_set)

        falseFunc = subSeq.evaluate(false_set)

        if trueFunc is False:
            newSeq = BoolSequence(operator="AND")
            if falseFunc is True:
                newSeq.append(-valname)
            elif falseFunc is False:
                newSeq.elements = False
                newSeq.level = -1
            else:
                newSeq.append(-valname, falseFunc)
                newSeq.join_operators(self_level=True)

            if updt:
                self.update(newSeq, pos)
            else:
                self.assign(newSeq)
            return True

        elif trueFunc is True:
            newSeq = BoolSequence(operator="OR")
            if falseFunc is True:
                newSeq.elements = True
                newSeq.level = -1
            elif falseFunc is False:
                newSeq.append(valname)
            else:
                newSeq.append(valname, falseFunc)
                newSeq.join_operators(self_level=True)

            if updt:
                self.update(newSeq, pos)
            else:
                self.assign(newSeq)
            return True

        if falseFunc is False:
            newSeq = BoolSequence(operator="AND")
            if trueFunc is True:
                newSeq.append(valname)
            elif trueFunc is False:
                newSeq.elements = False
                newSeq.level = -1
            else:
                newSeq.append(valname, trueFunc)
                newSeq.join_operators(self_level=True)
            if updt:
                self.update(newSeq, pos)
            else:
                self.assign(newSeq)
            return True

        elif falseFunc is True:
            newSeq = BoolSequence(operator="OR")
            if trueFunc is True:
                newSeq.elements = True
                newSeq.level = -1
            elif trueFunc is False:
                newSeq.append(-valname)
            else:
                newSeq.append(-valname, trueFunc)
                newSeq.join_operators(self_level=True)
            if updt:
                self.update(newSeq, pos)
            else:
                self.assign(newSeq)
            return True

    def evaluate(self, valueSet):
        """Return the result of the evaluation of the BoolSequence given the known values of the variables in "valueSet".
        Result can be a Boolean value or the reduced expresion of the BoolSequence."""
        if type(self.elements) is bool:
            return self.elements
        self.group_single()
        newSeq = self.copy()
        for name, value in valueSet.items():
            newSeq.substitute(name, value)
            if type(newSeq.elements) is bool:
                return newSeq.elements

        return newSeq.elements if type(newSeq.elements) is bool else newSeq

    def set_def(self, expression):
        """Set the expression of the Boolean function in the BoolSequence instance.
        "expresion" is the string object. The definition should have MCNP syntax cell definition.
        """
        terms, operator = outer_terms(expression)
        self.operator = operator
        self.level = 0
        lev0Seq = set()
        lev0SeqAbs = set()
        for t in terms:
            if is_integer(t):
                val = int(t.strip("(").strip(")"))
                lev0Seq.add(val)
                lev0SeqAbs.add(abs(val))
                # self.elements.append(int(t.strip('(').strip(')')))
            else:
                x = BoolSequence(t)
                self.level = max(x.level + 1, self.level)
                self.append(x)

        # check if in integer sequence there is surface sequence s -s
        if len(lev0Seq) != len(lev0SeqAbs):
            if self.operator == "AND":
                self.elements = False
            else:
                self.elements = True
            self.level = -1
        else:
            self.append(*lev0Seq)

        self.group_single()

    def group_single(self):
        """group integers found in Sequence with level > 1.
        (e.g. change AND[1 2 3 OR[2 4]] to AND[ AND[1 2 3] OR[2 3]] )."""
        if self.level == 0:
            return
        if type(self.elements) is bool:
            return
        group = []
        for e in reversed(self.elements):
            if type(e) is int:
                group.append(e)
                self.elements.remove(e)
            elif e.level == 0 and len(e.elements) == 1:
                group.append(e.elements[0])
                self.elements.remove(e)

        if not group:
            return
        seq = BoolSequence()
        seq.elements.extend(group)
        seq.operator = self.operator
        seq.level = 0
        self.elements.insert(0, seq)

    def get_surfaces_numbers(self):
        """Return the list of all surfaces in the BoolSequence definition."""
        if type(self.elements) is bool:
            return tuple()
        surf = set()
        for e in self.elements:
            if type(e) is int:
                surf.add(abs(e))
            else:
                surf.update(e.get_surfaces_numbers())
        return surf

    def level_update(self):
        """Update the level value of the BoolSequence."""
        if type(self.elements) is bool:
            self.level = 0
            return

        self.level = 0
        for e in self.elements:
            if type(e) is int:
                continue
            e.level_update()
            self.level = max(e.level + 1, self.level)


def insert_in_sequence(Seq, trgt, nsrf, operator):
    """Substitute the variable trgt by the sequence "(trgt:nsrg)" or "(trgt nsf)" in the
    BoolSequence Seq"""
    if operator == "OR":
        newSeq = BoolSequence(f"{trgt}:{nsrf}")
    else:
        newSeq = BoolSequence(f"{trgt} {nsrf}")

    substitute_integer_element(Seq, trgt, newSeq)
    Seq.level_update()
    # Seq.join_operators()


def substitute_integer_element(Seq, target, newElement):
    """Substitute the variable target by the sequence newElement in the
    BoolSequence Seq"""
    for i, e in enumerate(Seq.elements):
        if type(e) is int:
            if e == target:
                Seq.elements[i] = newElement
        else:
            substitute_integer_element(e, target, newElement)


def outer_terms(expression, value="number"):
    """Return the list and the boolean operator of the outter terms of the expression."""
    if value == "number":
        # reValue = number
        reValue = mix
        nullVal = "0"
    else:
        reValue = TFX
        nullVal = "o"

    expr = expression

    # Loop until no redundant parentheses are found
    cont = True

    while cont:
        # Loop over most inner parentheses
        pos = 0
        cont = False
        while True:
            m = mostinner.search(expr, pos)
            if not m:
                break
            cont = True
            if redundant(m, expr):
                # remove redundant parentheses
                expr = expr[: m.start()] + " " + expr[m.start() + 1 : m.end() - 1] + " " + expr[m.end() :]
            else:
                # replace no redundant parentheses by 0 and : by ;
                zeros = "[" + nullVal * (m.end() - m.start() - 2) + "]"
                expr = expr[: m.start()] + zeros + expr[m.end() :]

            pos = m.end()

    if ":" in expr:
        terms = []
        pos = 0
        while True:
            new_pos = expr.find(":", pos)
            if new_pos == -1:
                terms.append(expression[pos:].strip())
                break
            terms.append(expression[pos:new_pos].strip())
            pos = new_pos + 1
        return (terms, "OR")
    else:
        terms = []
        pos = 0
        while True:
            m = reValue.search(expr, pos)
            if not m:
                break
            terms.append(expression[m.start() : m.end()])
            pos = m.end()
        return (terms, "AND")


def redundant(m, geom):
    """Check if the inner parentheses are redundant."""
    term = m.group()

    # Find first valid character at the left of the  parenthese
    left_ok = True
    left = m.start() - 1
    while left > -1:
        if geom[left] in ("\n", "C", "$", " "):
            left -= 1
        else:
            if geom[left] not in ("(", ":"):
                left_ok = False
            break

    # check if no ':' (or) are inside the parenthese
    # if not, parentheses are redundants
    if term.find(":") == -1:
        return True

    # Find first valid character at the right of the  parenthese
    right_ok = True
    right = m.end()
    while right < len(geom):
        if geom[right] in ("\n", "C", "$", " "):
            right += 1
        else:
            if geom[right] not in (")", ":"):
                right_ok = False
            break

    # if parentheses are like:
    # {( or : } ( ....... ) {) or :}
    # parentheses are redundants

    if left_ok and right_ok:
        return True
    else:
        return False


def is_integer(x):
    try:
        int(x.strip("(").strip(")"))
        return True
    except:
        return False



import os
from pathlib import Path

import pytest

import geouned

path_to_cad = Path("testing/inputSTEP")
step_files = list(path_to_cad.rglob("*.stp")) + list(path_to_cad.rglob("*.step"))
# removing two geometries that are particularly slow to convert from CI testing
# these two geometries remain in the test suite for locally testing
if os.getenv("GITHUB_ACTIONS"):
    step_files.remove(Path("testing/inputSTEP/large/SCDR.stp"))
    step_files.remove(Path("testing/inputSTEP/large/Triangle.stp"))
suffixes = (".mcnp", ".xml", ".inp", ".py", ".serp")


@pytest.mark.parametrize("input_step_file", step_files)
def test_conversion(input_step_file):
    """Test that step files can be converted to openmc and mcnp files"""

    # sets up an output folder for the results
    output_dir = Path("tests_outputs") / input_step_file.with_suffix("")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_filename_stem = output_dir / input_step_file.stem

    # deletes the output MC files if they already exists
    for suffix in suffixes:
        output_filename_stem.with_suffix(suffix).unlink(missing_ok=True)

    my_options = geouned.Options(
        forceCylinder=False,
        newSplitPlane=True,
        delLastNumber=False,
        enlargeBox=2,
        nPlaneReverse=0,
        splitTolerance=0,
        scaleUp=True,
        quadricPY=False,
        Facets=False,
        prnt3PPlane=False,
        forceNoOverlap=False,
    )

    my_tolerances = geouned.Tolerances(
        relativeTol=False,
        relativePrecision=0.000001,
        value=0.000001,
        distance=0.0001,
        angle=0.0001,
        pln_distance=0.0001,
        pln_angle=0.0001,
        cyl_distance=0.0001,
        cyl_angle=0.0001,
        sph_distance=0.0001,
        kne_distance=0.0001,
        kne_angle=0.0001,
        tor_distance=0.0001,
        tor_angle=0.0001,
        min_area=0.01,
    )

    my_numeric_format = geouned.NumericFormat(
        P_abc="14.7e",
        P_d="14.7e",
        P_xyz="14.7e",
        S_r="14.7e",
        S_xyz="14.7e",
        C_r="12f",
        C_xyz="12f",
        K_xyz="13.6e",
        K_tan2="12f",
        T_r="14.7e",
        T_xyz="14.7e",
        GQ_1to6="18.15f",
        GQ_7to9="18.15f",
        GQ_10="18.15f",
    )

    my_settings = geouned.Settings(
        matFile="",
        voidGen=True,
        debug=False,
        compSolids=False,
        simplify="no",
        exportSolids="",
        minVoidSize=200.0,  # units mm
        maxSurf=50,
        maxBracket=30,
        voidMat=[],
        voidExclude=[],
        startCell=1,
        startSurf=1,
        sort_enclosure=False,
    )

    geo = geouned.CadToCsg(
        options=my_options,
        settings=my_settings,
        tolerances=my_tolerances,
        numeric_format=my_numeric_format,
    )

    geo.load_step_file(filename=f"{input_step_file.resolve()}", skip_solids=[])

    geo.start()

    geo.export_csg(
        title="Converted with GEOUNED",
        geometryName=f"{output_filename_stem.resolve()}",
        outFormat=(
            "openmc_xml",
            "openmc_py",
            "serpent",
            "phits",
            "mcnp",
        ),
        volSDEF=True,  # changed from the default
        volCARD=False,  # changed from the default
        UCARD=None,
        dummyMat=True,  # changed from the default
        cellCommentFile=False,
        cellSummaryFile=False,  # changed from the default
    )

    for suffix in suffixes:
        assert output_filename_stem.with_suffix(suffix).exists()


@pytest.mark.parametrize(
    "input_json_file",
    ["tests/config_cadtocsg_complete_defaults.json", "tests/config_cadtocsg_minimal.json"],
)
def test_cad_to_csg_from_json_with_defaults(input_json_file):

    # deletes the output MC files if they already exists
    for suffix in suffixes:
        Path("csg").with_suffix(suffix).unlink(missing_ok=True)

    my_cad_to_csg = geouned.CadToCsg.from_json(input_json_file)
    assert isinstance(my_cad_to_csg, geouned.CadToCsg)

    assert my_cad_to_csg.filename == "testing/inputSTEP/BC.stp"
    assert my_cad_to_csg.options.forceCylinder == False
    assert my_cad_to_csg.tolerances.relativeTol == False
    assert my_cad_to_csg.numeric_format.P_abc == "14.7e"
    assert my_cad_to_csg.settings.matFile == ""

    for suffix in suffixes:
        assert Path("csg").with_suffix(suffix).exists()


def test_cad_to_csg_from_json_with_non_defaults():

    # deletes the output MC files if they already exists
    for suffix in suffixes:
        Path("csg").with_suffix(suffix).unlink(missing_ok=True)

    my_cad_to_csg = geouned.CadToCsg.from_json("tests/config_cadtocsg_non_defaults.json")
    assert isinstance(my_cad_to_csg, geouned.CadToCsg)

    assert my_cad_to_csg.filename == "testing/inputSTEP/BC.stp"
    assert my_cad_to_csg.options.forceCylinder == True
    assert my_cad_to_csg.tolerances.relativePrecision == 2e-6
    assert my_cad_to_csg.numeric_format.P_abc == "15.7e"
    assert my_cad_to_csg.settings.matFile == "non default"

    for suffix in suffixes:
        assert Path("csg").with_suffix(suffix).exists()


def test_writing_to_new_folders():
    """Checks that a folder is created prior to writing output files"""

    geo = geouned.CadToCsg()
    geo.load_step_file(filename="testing/inputSTEP/BC.stp", skip_solids=[])
    geo.start()

    for outformat in ["mcnp", "phits", "serpent", "openmc_xml", "openmc_py"]:
        geo.export_csg(
            geometryName=f"tests_outputs/new_folder_for_testing_{outformat}/csg",
            cellCommentFile=False,
            cellSummaryFile=False,
            outFormat=[outformat],
        )
        geo.export_csg(
            geometryName=f"tests_outputs/new_folder_for_testing_{outformat}_cell_comment/csg",
            cellCommentFile=True,
            cellSummaryFile=False,
            outFormat=[outformat],
        )
        geo.export_csg(
            geometryName=f"tests_outputs/new_folder_for_testing_{outformat}_cell_summary/csg",
            cellCommentFile=False,
            cellSummaryFile=True,
            outFormat=[outformat],
        )


def test_with_relative_tol_true():

    # test to protect against incorrect attribute usage in FreeCAD
    # more details https://github.com/GEOUNED-org/GEOUNED/issues/154

    geo = geouned.CadToCsg(
        tolerances=geouned.Tolerances(relativeTol=False),
    )
    geo.load_step_file(filename=f"{step_files[1].resolve()}", skip_solids=[])
    geo.start()

    geo = geouned.CadToCsg(
        tolerances=geouned.Tolerances(relativeTol=True),
    )
    geo.load_step_file(filename=f"{step_files[1].resolve()}", skip_solids=[])
    geo.start()



import re

#########################################
# define patterns to be found in string #
#########################################


# used in card_split
celmat = re.compile(
    r"(?P<cel>^ *\d+) +(?P<scnd>(\d+|like))", re.I
)  # identify first two entries on cell card (cell name, material)
grlnumber = re.compile(
    r"[-+]?(\d+\.\d+|\.\d+|\d+\.?)(e[+-]\d+)?", re.I
)  # identify a general number form signed integer, float or exponential
param = re.compile(
    r"((^|\n {5})[\(\):\-\+\d+\.\# ]*)([\*a-z])", re.I
)  # identity begining of the paramter part of the cell card
likebut = re.compile(r"but", re.I)  # identify likebut  card
trans = re.compile(r"trcl|fill= *\d+[ c\$\n]*\(,", re.I)  # identify tranformed card

# user in get_stat function
reword = re.compile(r"(\d+|\(|\)|:|\#)")  # word identification in cell line
compcell = re.compile(r"\#\d")  # identify hashcell complement operator

# used in Complementary operator function
number = re.compile(r"(?P<number>[-+]?\d+)")  # signed (+-) (or not) numbers
# leftp=re.compile(r"^ *(?P<left>[-\d\(\#])",re.M)                               # identify first valid character
leftp = re.compile(r"(?P<left>[-+\d\(\#])", re.I)  # identify first valid character
rightp = re.compile(r"(?P<right>[ c\$\n]*$)", re.I)  # identify last valid character
# interblk=re.compile(r"(?P<previous>\d)(?P<next>(( +| *(\$)?\n(C\n)* *)[-+]?\d))") # two numbers separated by blank (or newline or comments)
# intercls=re.compile(r"(?P<previous>\))(?P<next>(( *| *(\$)?\n(C\n)* *)[-+]?\d))") # closed parenthesis followed by number
# interopn=re.compile(r"(?P<previous>\d)(?P<next>(( *| *(\$)?\n(C\n)* *)\())")   # number followed by opened parenthesis
# intercop=re.compile(r"(?P<previous>\))(?P<next>(( *| *(\$)?\n(C\n)* *)\())")   # closed parenthesis followed by opened parenthesis
interblk = re.compile(
    r"(?P<previous>\d)(?P<next>(( +| *((\n *)?\$|\nC)*\n *)[-+]?\d))"
)  # two numbers separated by blank (or newline or comments)
intercls = re.compile(
    r"(?P<previous>\))(?P<next>(( *| *((\n *)?\$|\nC)*\n *)[-+]?\d))"
)  # closed parenthesis followed by number
interopn = re.compile(r"(?P<previous>\d)(?P<next>(( *| *((\n *)?\$|\nC)*\n *)\())")  # number followed by opened parenthesis
intercop = re.compile(
    r"(?P<previous>\))(?P<next>(( *| *((\n *)?\$|\nC)*\n *)\())"
)  # closed parenthesis followed by opened parenthesis
colonamp = re.compile(r"[:&]")  # colon or ampersand

# used for remove redundant parenthesis function
mostinner = re.compile(r"\([^\(^\)]*\)")  # identify most inner parentheses
bracketsemi = re.compile(r"[\]\[;]")  # square bracket or semicolon
blnkline = re.compile(r"^ *\n", re.M)  # identify blank line
contline = re.compile(r"\n {0,4}(?P<start>[^c^ ])", re.I)  # identify character other than 'C' in fisrt 5 columns
comdollar = re.compile(r"\n(?P<blnk> *)\$")  # identify dollar on 'blank line'
startgeom = re.compile(r"(?P<previous>^ *)(?P<start>[\-\+\d])")  # identify beginning of the geomtric part
endgeom = re.compile(r"(?P<last>\d)(?P<next> *((\n *)?\$|\nc)?(\n *)?$)", re.I)  # identify end of the geomtric part
# endgeom=re.compile(r"(?P<last>\d)(?P<next> *(\$|\nc)?(\n *)?$)",re.I)                      # identify end of the geomtric part

# other
rehash = re.compile(r"# *(\d+|\()")  # find beginning of complementary operator (both cell and surf)
parent = re.compile(r"[\(|\)]")  # position of open and close parenthesis (get_hashcell)
gline = re.compile(
    r"(^ ?[\(\):\-\+\d+\.\# ]+|\n {5}[\(\):\-\+\d+\.\# ]+)", re.I
)  # valid geometric part of the line       (remove/restore_comments)
comments = re.compile(r"((\n *)?\$|\n *c)", re.I)  # begining of comment part               (remove/restore_comments)
# comments=re.compile(r"\$|\n *c",re.I)                               # begining of comment part               (remove/restore_comments)
lastCR = re.compile(r"\n *$")  # last newline "\n" on line string


# restore curve parentheses and colon characters
def reverse_repl(m):
    symb = m.group(0)
    if symb == "[":
        return "("
    elif symb == "]":
        return ")"
    else:
        return ":"


def countP(string):
    lp = string.count("(")
    rp = string.count(")")
    return (lp, rp)


def redundant(m, geom):
    """check if the inner parentheses are redundant"""
    term = m.group()

    # Find first valid character at the left of the  parenthese
    leftOK = True
    left = m.start() - 1
    while left > -1:
        if geom[left] in ("\n", "C", "$", " "):
            left -= 1
        else:
            if geom[left] not in ("(", ":"):
                leftOK = False
            break

    # check if no ':' (or) are inside the parenthese
    # if not, parentheses are redundants
    if term.find(":") == -1:
        return True

    # Find first valid character at the right of the  parenthese
    rightOK = True
    right = m.end()
    while right < len(geom):
        if geom[right] in ("\n", "C", "$", " "):
            right += 1
        else:
            if geom[right] not in (")", ":"):
                rightOK = False
            break

    # if parentheses are like:
    # {( or : } ( ....... ) {) or :}
    # parentheses are redundants

    if leftOK and rightOK:
        return True
    else:
        return False


def remove_redundant(geom):
    """return cell without redundant parenthesis"""

    # Loop until no redundant parentheses are found
    cont = True
    while cont:
        # Loop over most inner parentheses
        pos = 0
        cont = False
        while True:
            m = mostinner.search(geom, pos)
            if not m:
                break
            cont = True
            if redundant(m, geom):
                # remove redundant parentheses
                geom = geom[: m.start()] + " " + geom[m.start() + 1 : m.end() - 1] + " " + geom[m.end() :]
            else:
                # replace no redundant parentheses by [] and : by ;
                term = geom[m.start() + 1 : m.end() - 1].replace(":", ";")
                geom = geom[: m.start()] + "[" + term + "]" + geom[m.end() :]
            pos = m.end()

    # restore curved parenthesis and colon
    geom = re.sub(bracketsemi, reverse_repl, geom)

    # remove possible blank line
    geom = re.sub(blnkline, "", geom)

    # ensure 5 blanks continous line
    geom = re.sub(contline, r"\n     \g<start>", geom)

    # remove last newline at the end of the line string
    geom = re.sub(lastCR, "", geom)

    return geom



##############################
# Module to write PHITS input #
##############################

# Mainly based on the MCNPFormat.py
# Some are modified for PHITS
# 1. Chosen the graveyard surface as the boundary of the outer void region (-1)
# 2. Unified the auto genarated void regions when envelope*_*_ and/or enclosure*_*_ are not setted
# 3. Applied MATERIAL DEFINITION block when dummyMat = True
# 4. Applied VOLUME DEFINITION block when volCARD = True
# 5. Eliminated the only MCNP related parts
# 6. Added some comments to remind

import logging
import re
from datetime import datetime
from pathlib import Path
from importlib.metadata import version

import FreeCAD

from ..utils.basic_functions_part1 import is_opposite, points_to_coeffs
from ..utils.functions import SurfacesDict
from .functions import (
    CellString,
    change_surf_sign,
    phits_surface,
    write_phits_cell_def,
)

logger = logging.getLogger("general_logger")


class PhitsInput:
    def __init__(
        self,
        Meta,
        Surfaces,
        options,
        tolerances,
        numeric_format,
        title,
        volSDEF,
        volCARD,
        UCARD,
        dummyMat,
        step_filename,
        matFile,
        voidMat,
        startCell,
    ):
        self.Title = title
        self.VolSDEF = volSDEF
        self.VolCARD = volCARD
        self.U0CARD = UCARD
        self.DummyMat = dummyMat
        self.Matfile = matFile
        self.voidMat = voidMat
        self.startCell = startCell
        self.Cells = Meta
        self.tolerances = tolerances
        self.numeric_format = numeric_format
        self.options = options
        self.Options = {"Volume": self.VolCARD, "Universe": self.U0CARD}

        self.step_filename = step_filename
        if isinstance(self.step_filename, (tuple, list)):
            self.step_filename = "; ".join(self.step_filename)

        if self.Title == "":
            self.Title = self.step_filename

        self.get_surface_table()
        self.simplify_planes(Surfaces)

        self.Surfaces = self.sorted_surfaces(Surfaces)
        self.Materials = set()

        return

    def write_phits(self, filename):
        logger.info(f"write PHITS file {filename}")
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(file=filename, mode="w", encoding="utf-8") as self.inpfile:
            self.write_phits_header()

            cHeader = """\
$
$ ##########################################################
$                  CELL DEFINITION
$ ##########################################################
$
[CELL]\n"""

            self.inpfile.write(cHeader)
            self.write_phits_cell_block()
            self.inpfile.write(" \n")

            surfaceHeader = """\
$
$ ##########################################################
$                  SURFACE DEFINITION
$ ##########################################################
$
[SURFACE]\n"""

            self.inpfile.write(surfaceHeader)
            self.write_phits_surface_block()
            self.inpfile.write(" \n")

            materialHeader = """\
$
$ ##########################################################
$                  MATERIAL DEFINITION
$ ##########################################################
$ All material labels present in this model are listed below
$ Need to change the dummy material definition(H2O1) to appropriate one(s) 
$
[MATERIAL]\n"""

            if self.DummyMat:
                self.inpfile.write(materialHeader)
                self.write_phits_source_block()
                self.inpfile.write(" \n")

            volHeader = """\
$
$ ##########################################################
$                  VOLUME DEFINITION
$ ##########################################################
$ The CAD calculated volume(s) is/are quoted for solid cell(s).
$ Note that the auto-generated void volumes are not calculated,
$ set all at 1.0cm3 tentatively.
$
[VOLUME] off\n"""

            if self.Options["Volume"]:
                self.inpfile.write(volHeader)
                self.write_phits_volume_block()
                self.inpfile.write(" \n")

            return

    def write_phits_header(self):

        freeCAD_Version = "{V[0]:}.{V[1]:}.{V[2]:}".format(V=FreeCAD.Version())

        Header = (
            "$ "
            """{}
$   ______ _______  _____      _     _ __   _ _______ ______  
$  |  ____ |______ |     | ___ |     | | \  | |______ |     \ 
$  |_____| |______ |_____|     |_____| |  \_| |______ |_____/
$ Version : {}
$ FreeCAD Version : {} 
$ PHITSFormat Version :  0.0.2.3     06/03/2024\n""".format(
                self.Title, version("geouned"), freeCAD_Version
            )
        )

        Information = f"""$
$ *************************************************************
$ Original Step file : {self.step_filename}
$
$ Creation Date : {datetime.now()}
$ Solid Cells   : {self.__solidCells__}
$ Total Cells   : {self.__cells__}
$ Surfaces      : {len(self.Surfaces)}
$ Materials     : {len(self.__materials__)}
$
$ **************************************************************


"""

        self.inpfile.write(Header)
        self.inpfile.write(Information)
        return

    def write_phits_cell_block(self):

        enclenvChk = []
        enclenvChk = self.step_filename_label_chk(self.step_filename)

        if enclenvChk:
            logger.info("Unified the inner void cell(s) definition")
            for i, cell in enumerate(self.Cells):
                self.write_phits_cells_uni_void_def(cell)
            return
        else:
            for i, cell in enumerate(self.Cells):
                self.write_phits_cells(cell)
            return

    def step_filename_label_chk(self, filename):

        enclenvList = []
        with open(filename) as f:
            enclenvList = f.readlines()
        enclLabel = re.search("enclosure(?P<encl>[0-9]+)_(?P<parent>[0-9]+)_", str(enclenvList))
        envelLabel = re.search("envelope(?P<env>[0-9]+)_(?P<parent>[0-9]+)_", str(enclenvList))
        cond1 = enclLabel == None
        cond2 = envelLabel == None
        return cond1 and cond2

    def write_phits_surface_block(self):

        for surf in self.Surfaces:
            self.write_phits_surfaces(surf)

    def write_phits_cells(self, cell):

        index = cell.label

        # if index is None objet not contain cell definition
        # but a comment to insert between cells
        if cell.__id__ is None:
            comment = self.comment_line(cell.Comments)
            self.inpfile.write(comment)
            return

        if cell.Material == 0:
            if cell.MatInfo == "Graveyard":
                cell.MatInfo = "Outer void"
                cellHeader = f"{index:<5d} {-1:<5d}  "

            elif cell.MatInfo == "Graveyard_in":
                cell.MatInfo = "Inner void"
                if self.voidMat != []:
                    self.Materials.add(self.voidMat[0])
                    if abs(self.voidMat[1]) < 1e-2:
                        cellHeader = "{:<5d} {:<5d} {:11.4e} ".format(index, self.voidMat[0], self.voidMat[1])
                    else:
                        cellHeader = "{:<5d} {:<5d} {:11.7f} ".format(index, self.voidMat[0], self.voidMat[1])
                else:
                    cellHeader = f"{index:<5d} {0:<5d}  "

            else:
                cellHeader = f"{index:<5d} {0:<5d}  "

        else:
            self.Materials.add(cell.Material)
            if self.Matfile == "" and cell.EnclosureID != 0:
                cellHeader = f"{index:<5d} {cell.Material:<5d} c{cell.Material:<5d} "
            else:
                if abs(cell.Density) < 1e-2:
                    cellHeader = "{:<5d} {:<5d} {:11.4e} ".format(index, cell.Material, cell.Density)
                else:
                    cellHeader = "{:<5d} {:<5d} {:11.7f} ".format(index, cell.Material, cell.Density)

        phitscell = "{}{}\n{}{}".format(
            cellHeader,
            self.cell_format(cell.Definition, offset=len(cellHeader)),
            self.option_format(cell),
            self.comment_format(cell.Comments, cell.MatInfo),
        )
        self.inpfile.write(phitscell)
        return

    def write_phits_cells_uni_void_def(self, cell):

        index = cell.label

        # if index is None objet not contain cell definition
        # but a comment to insert between cells
        if cell.__id__ is None:
            comment = self.comment_line(cell.Comments)
            self.inpfile.write(comment)
            return
        """
        # Graveyard_in and Graveyard is changed.
        # Grayveyard sphere is redefined as the boundary surface 
        # of the inner and outer void region.  
        # Auto-generated void cell(s) is/are eliminated.
        # To exclude solid cell(s) from the inner void's defenition,
        # a string of '#(Solid Cell No.)', or inclsolidCells, 
        # is appended to the new inner void cell definition
        # after self.cell_format(cell.Definition) process.       
        """

        if cell.Void:
            if cell.MatInfo == "Graveyard_in":
                cell.MatInfo = "Inner void"

                newInnerVoidCell = cell.Definition.elements[:1]
                cell.Definition.operator = "AND"
                cell.Definition.elements = newInnerVoidCell

                inclSolidCells = ""
                startVoidIndex = self.__solidCells__ + self.startCell
                eliminated_endVoidIndex = self.__cells__ + self.startCell - 3

                if self.startCell == startVoidIndex - 1:
                    inclSolidCells = f"{'':1s}#{self.startCell}"
                else:
                    for i in range(self.startCell, startVoidIndex):
                        inclSolidCells += f"{'':1s}#{i}"

                if startVoidIndex == eliminated_endVoidIndex:
                    one_mervoid_str = "VOID CELL {} merged, so the auto-genarated void definition is eliminated\n"
                    cell.Comments = one_mervoid_str.format(startVoidIndex)
                else:
                    some_mervoid_str = "VOID CELLs {}-{} merged, so the auto-genarated void definitions are eliminated\n"
                    cell.Comments = some_mervoid_str.format(startVoidIndex, eliminated_endVoidIndex)
                if self.voidMat != []:
                    self.Materials.add(self.voidMat[0])
                    if abs(self.voidMat[1]) < 1e-2:
                        cellHeader = "{:<5d} {:<5d} {:11.4e} ".format(index, self.voidMat[0], self.voidMat[1])
                    else:
                        cellHeader = "{:<5d} {:<5d} {:11.7f} ".format(index, self.voidMat[0], self.voidMat[1])
                else:
                    cellHeader = f"{index:<5d} {0:<5d}  "

                phitscell = "{}{}\n{}{}\n".format(
                    cellHeader,
                    self.inner_void_def(inclSolidCells, cell.Definition, offset=len(cellHeader)),
                    self.option_format(cell),
                    self.comment_format(cell.Comments, cell.MatInfo),
                )
                self.inpfile.write(phitscell)
                return

            elif cell.MatInfo == "Graveyard":
                cellHeader = f"{index:<5d} {-1:<5d}  "
                cell.MatInfo = "Outer void"

            else:
                return

            """
        # To check auto-generated voids, apply this commented out section instead
        # and comment out above from "if cell.Void:..." to "... else: return"
        # In addition, if you set volCARD = True and want for all void regions to come apperes in [VOLUME],
        # comment out some part in the def write_phits_volume_block() section also.       
        if cell.Material == 0:
            logger.info(cell.IsEnclosure)
            if cell.MatInfo == 'Graveyard':
                cellHeader = '{:<5d} {:<5d}  '.format(index,-1)
            else:
                cellHeader = '{:<5d} {:<5d}  '.format(index,0)
           """

        else:
            self.Materials.add(cell.Material)
            if self.Matfile == "" and cell.EnclosureID != 0:
                cellHeader = f"{index:<5d} {cell.Material:<5d} c{cell.Material:<5d} "
            else:
                if abs(cell.Density) < 1e-2:
                    cellHeader = "{:<5d} {:<5d} {:11.4e} ".format(index, cell.Material, cell.Density)
                else:
                    cellHeader = "{:<5d} {:<5d} {:11.7f} ".format(index, cell.Material, cell.Density)

        phitscell = "{}{}\n{}{}".format(
            cellHeader,
            self.cell_format(cell.Definition, offset=len(cellHeader)),
            self.option_format(cell),
            self.comment_format(cell.Comments, cell.MatInfo),
        )
        self.inpfile.write(phitscell)
        return

    def write_phits_surfaces(self, surface):
        """Write the surfaces in PHITS format"""

        PHITS_def = phits_surface(
            surface.Index,
            surface.Type,
            surface.Surf,
            self.options,
            self.tolerances,
            self.numeric_format,
        )
        if PHITS_def:
            PHITS_def += "\n"
            self.inpfile.write(PHITS_def)
        else:
            logger.info(f"Surface {surface.Type} cannot be written in PHITS input")
        return

    def write_phits_source_block(self):

        if self.DummyMat:
            mat = list(self.Materials)
            mat.sort()
            MATID = []
            MATCARD = ""
            for i, cell in enumerate(self.Cells):
                if (cell.Material in mat) and (cell.Material not in MATID):
                    MATID.append(cell.Material)
                    if self.Matfile == "" and cell.EnclosureID != 0:
                        mismat_comment = "$ Change dummyMat M{}, {} c{} g/cm3 is assigned\n M{:<6d} H 2 O 1\n"
                        MATCARD += mismat_comment.format(cell.Material, cell.MatInfo, cell.Material, cell.Material)
                    else:
                        mat_comment = "$ Change dummyMat M{} to {}, Density = {}g/cm3\n M{:<6d} H 2 O 1\n"
                        MATCARD += mat_comment.format(
                            cell.Material,
                            cell.MatInfo,
                            -1 * cell.Density,
                            cell.Material,
                        )
            Block = MATCARD + "\n"

        self.inpfile.write(Block)

    def write_phits_volume_block(self):

        vol = f"{'':5s}reg{'':5s}vol\n"

        startVoidIndex = self.__solidCells__ + self.startCell
        eliminated_endVoidIndex = self.__cells__ + self.startCell - 3

        enclenvChk = []
        enclenvChk = self.step_filename_label_chk(self.step_filename)

        if enclenvChk and self.Options["Volume"]:
            for i, cell in enumerate(self.Cells):
                if cell.__id__ is not None:
                    if cell.Void and startVoidIndex == eliminated_endVoidIndex:
                        if cell.label == startVoidIndex:
                            logger.info(f"Eliminated the merged void cell {cell.label} from [VOLUME] section")
                        else:
                            vol += f"{'':6s}{cell.label}{'':6s}1.0\n"
                    elif cell.Void:
                        if cell.label in range(startVoidIndex, eliminated_endVoidIndex + 1):
                            logger.info(f"Eliminated the merged void cell {cell.label} from [VOLUME] section")
                        else:
                            vol += f"{'':6s}{cell.label}{'':6s}1.0\n"
                    else:
                        vol += f"{'':6s}{cell.label}{'':6s}{cell.Volume * 0.001:6e}\n"
        else:
            if self.Options["Volume"]:
                for i, cell in enumerate(self.Cells):
                    if cell.__id__ is not None:
                        if cell.Void:
                            vol += f"{'':6s}{cell.label}{'':6s}1.0\n"
                        else:
                            vol += "{:6s}{}{:6s}{:6e}\n".format("", cell.label, "", cell.Volume * 1e-3)

        self.inpfile.write(vol)

        """
        # If you want for all eliminated the merged void cells to come apperes in [VOLUME],
        # apply this commented out section instead
        # and comment out above from "startVoidIndex =..." to "... self.inpfile.write(vol)"
        if self.Options['Volume']:
            for i,cell in enumerate(self.Cells):
                if cell.__id__ is not None :
                    if cell.Void : 
                        vol +='{:6s}{}{:6s}1.0\n'.format('',cell.label,'')
                    else:
                        vol +='{:6s}{}{:6s}{:e}\n'.format('',cell.label,'',cell.Volume*1e-3)

        self.inpfile.write(vol)
        """

    def cell_format(self, Definition, offset=11):
        return write_phits_cell_def(Definition, tabspace=11, offset=offset)

    def inner_void_def(self, innerSolidCells, Definition, offset=11):
        newInnerVoidDef = self.cell_format(Definition, offset)
        strdef = CellString(tabspace=11)
        strdef.add(newInnerVoidDef + innerSolidCells)
        strdef.wrap_line(offset)
        return strdef.str

    def option_format(self, cell):

        option = ""
        if self.Options["Volume"]:
            if not cell.Void:
                option = f"${'':11s}Vol={cell.Volume * 0.001:e} cm3\n"
            else:
                option = f"${'':11s}Vol=1.0 cm3\n"

        if self.Options["Universe"] is not None:
            option += f"{'':11s}U={self.Options['Universe']}\n"

        return option

    def comment_format(self, cComment, mComment=None):

        comment = ""
        if mComment:
            mComment = mComment.split("\n")
            for c in mComment:
                if c:
                    comment += f"{'':11s}${c}\n"

        if cComment.strip() != "":
            cComment = cComment.strip().split("\n")
            for c in cComment:
                if c:
                    comment += f"{'':11s}${c}\n"
        return comment

    def comment_line(self, lineComment):
        lineComment = lineComment.strip().split("\n")
        comment = ""
        if lineComment:
            comment = "$ \n"
            for c in lineComment:
                if c:
                    comment += f"$ {c}\n"
            comment += "$ \n"
        return comment

    def get_surface_table(self):
        self.surfaceTable = {}
        self.__solidCells__ = 0
        self.__cells__ = 0
        self.__materials__ = set()

        for i, CellObj in enumerate(self.Cells):
            if CellObj.__id__ is None:
                continue
            self.__cells__ += 1
            if CellObj.Material != 0:
                self.__materials__.add(CellObj.Material)

            surf = CellObj.Definition.get_surfaces_numbers()
            if not CellObj.Void:
                self.__solidCells__ += 1
            for index in surf:
                if index in self.surfaceTable.keys():
                    self.surfaceTable[index].add(i)
                else:
                    self.surfaceTable[index] = {i}
        return

    def simplify_planes(self, Surfaces):

        for p in Surfaces["PX"]:
            if p.Surf.Axis[0] < 0:
                p.Surf.Axis = FreeCAD.Vector(1, 0, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PY"]:
            if p.Surf.Axis[1] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 1, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PZ"]:
            if p.Surf.Axis[2] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 0, 1)
                self.change_surf_sign(p)

        if self.options.prnt3PPlane:
            for p in Surfaces["P"]:
                if p.Surf.pointDef:
                    axis, d = points_to_coeffs(p.Surf.Points)
                    if is_opposite(axis, p.Surf.Axis):
                        self.change_surf_sign(p)
        return

    def sorted_surfaces(self, Surfaces):
        temp = SurfacesDict(Surfaces)
        surfList = []
        for ind in range(Surfaces.IndexOffset, Surfaces.surfaceNumber + Surfaces.IndexOffset):
            s = temp.get_surface(ind + 1)
            if s is not None:
                surfList.append(s)
                temp.del_surface(ind + 1)
        return surfList

    def change_surf_sign(self, p):

        if p.Index not in self.surfaceTable.keys():
            logger.info(f"{p.Type} Surface {p.Index} not used in cell definition) {p.Surf.Axis} {p.Surf.Position}")
            return

        for ic in self.surfaceTable[p.Index]:
            surf = self.Cells[ic].Definition.get_surfaces_numbers()
            for s in surf:
                if s == p.Index:
                    change_surf_sign(s, self.Cells[ic].Definition)

    def get_solid_cell_volume(self):

        solidList = []
        volumeList = []
        for m in self.Cells:
            if m.CellType == "solid" and m.__id__ is not None:
                solidList.append(m.label)
                volumeList.append(m.Volume * 1e-3)
        return solidList, volumeList



from . import additional_files as OutFiles
from .mcnp_format import McnpInput
from .openmc_format import OpenmcInput
from .phits_format import PhitsInput
from .serpent_format import SerpentInput


def write_geometry(
    UniverseBox,
    MetaList,
    Surfaces,
    settings,
    options,
    tolerances,
    numeric_format,
    geometryName,
    outFormat,
    cellCommentFile,
    cellSummaryFile,
    title,
    volSDEF,
    volCARD,
    UCARD,
    dummyMat,
    step_filename,
):

    supported_mc_codes = ("mcnp", "openmc_xml", "openmc_py", "serpent", "phits")
    for out_format in outFormat:
        if out_format not in supported_mc_codes:
            msg = f"outFormat {out_format} not in supported MC codes ({supported_mc_codes})"
            raise ValueError(msg)

    # write cells comments in file
    if cellCommentFile:
        OutFiles.comments_write(geometryName, MetaList)
    if cellSummaryFile:
        OutFiles.summary_write(geometryName, MetaList)

    if "mcnp" in outFormat:
        mcnpFilename = geometryName + ".mcnp"
        outBox = (
            UniverseBox.XMin,
            UniverseBox.XMax,
            UniverseBox.YMin,
            UniverseBox.YMax,
            UniverseBox.ZMin,
            UniverseBox.ZMax,
        )
        if settings.voidGen:
            outSphere = (Surfaces["Sph"][-1].Index, Surfaces["Sph"][-1].Surf.Radius)
        else:
            outSphere = None

        MCNPfile = McnpInput(
            MetaList,
            Surfaces,
            options,
            tolerances,
            numeric_format,
            title,
            volSDEF,
            volCARD,
            UCARD,
            dummyMat,
            step_filename,
        )
        MCNPfile.set_sdef((outSphere, outBox))
        MCNPfile.write_input(mcnpFilename)

    if "openmc_xml" in outFormat or "openmc_py" in outFormat:
        OMCFile = OpenmcInput(MetaList, Surfaces, options, tolerances, numeric_format)

    if "openmc_xml" in outFormat:
        omcFilename = geometryName + ".xml"
        OMCFile.write_xml(omcFilename)

    if "openmc_py" in outFormat:
        omcFilename = geometryName + ".py"
        OMCFile.write_py(omcFilename)

    if "serpent" in outFormat:
        serpentFilename = geometryName + ".serp"
        outBox = (
            UniverseBox.XMin,
            UniverseBox.XMax,
            UniverseBox.YMin,
            UniverseBox.YMax,
            UniverseBox.ZMin,
            UniverseBox.ZMax,
        )
        if settings.voidGen:
            outSphere = (Surfaces["Sph"][-1].Index, Surfaces["Sph"][-1].Surf.Radius)
        else:
            outSphere = None

        Serpentfile = SerpentInput(
            MetaList,
            Surfaces,
            options,
            tolerances,
            numeric_format,
            title,
            volSDEF,
            volCARD,
            UCARD,
            dummyMat,
            step_filename,
        )
        # Serpentfile.set_sdef((outSphere,outBox))
        Serpentfile.write_input(serpentFilename)

    if "phits" in outFormat:
        phitsFilename = geometryName + ".inp"
        PHITS_outBox = (
            UniverseBox.XMin,
            UniverseBox.XMax,
            UniverseBox.YMin,
            UniverseBox.YMax,
            UniverseBox.ZMin,
            UniverseBox.ZMax,
        )
        if settings.voidGen:
            PHITS_outSphere = (
                Surfaces["Sph"][-1].Index,
                Surfaces["Sph"][-1].Surf.Radius,
            )
        else:
            PHITS_outSphere = None

        PHITSfile = PhitsInput(
            MetaList,
            Surfaces,
            options,
            tolerances,
            numeric_format,
            title,
            volSDEF,
            volCARD,
            UCARD,
            dummyMat,
            step_filename,
            matFile=settings.matFile,
            voidMat=settings.voidMat,
            startCell=settings.startCell,
        )
        # PHITSfile.setSDEF_PHITS((PHITS_outSphere,PHITS_outBox))
        PHITSfile.write_phits(phitsFilename)



import typing
from numbers import Real


class Options:
    """A class for containing conversion options

    Args:
        forceCylinder (bool, optional): Use cylinder (instead of cones) as
            ancillary surface where unclosed torus surfaces are involved in
            the solid definition. Defaults to False.
        newSplitPlane (bool, optional): New method to consider plane as
            cutting surface during the decomposition process. Former method
            split first planes perpendicular to X,Y,Z axis and then the
            other planes involved in the solid definition. New method group
            all parallel planes independently whether their normal are
            along X,Y,Z axes, and start the decomposition process cutting
            first with the group having the highest number of parallel
            planes. Defaults to True.
        delLastNumber (bool, optional): Deleting the last word in the
            comment if it is a number. Defaults to False.
        enlargeBox (Real, optional): Enlarge box boundary when evaluating
            the constraint table during the simplification of the void cell
            definition. (unit is millimeter). Defaults to 2.
        nPlaneReverse (int, optional): Threshold value to determine whether
            cut with parallel planes should be carried out first. Defaults
            to 0.
        splitTolerance (Real, optional): Fuzzy tolerance value used in the
            FreeCAD function “BOPTools.SplitAPI.slice”. This function is
            used during the solid decomposition process. Defaults to 0.
        scaleUp (bool, optional): Scale up Fuzzy tolerance once get below
            1e-12. Defaults to True.
        quadricPY (bool, optional): In openMC python script format, the
            cones or cylinders no aligned with the X,Y, or Z axis can be
            defined using the openmc.Cone or open.Cylinder methods but can
            also be defined with their quadric parameter. If “quadricPY” is
            11 True then all cones and cylinders will be defined in the
            openMC python script format under their quadric form. Defaults
            to False.
        Facets (bool, optional): use alternative conversion module when
            geometry is defined by cells compound by only triangular plane
            faces. Defaults to False.
        prnt3PPlane (bool, optional): print 3 point plane definition in
            output as 3 points coordinates. Defaults to False.
        forceNoOverlap (bool, optional): force no overlaping cell
            definition. Adjacent cell definition are rested from current
            cell definition. Defaults to False.
    """

    def __init__(
        self,
        forceCylinder: bool = False,
        newSplitPlane: bool = True,
        delLastNumber: bool = False,
        enlargeBox: Real = 2.0,
        nPlaneReverse: int = 0,
        splitTolerance: Real = 0.0,
        scaleUp: bool = True,
        quadricPY: bool = False,
        Facets: bool = False,
        prnt3PPlane: bool = False,
        forceNoOverlap: bool = False,
    ):

        self.forceCylinder = forceCylinder
        self.newSplitPlane = newSplitPlane
        self.delLastNumber = delLastNumber
        self.enlargeBox = enlargeBox
        self.nPlaneReverse = nPlaneReverse
        self.splitTolerance = splitTolerance
        self.scaleUp = scaleUp
        self.quadricPY = quadricPY
        self.Facets = Facets
        self.prnt3PPlane = prnt3PPlane
        self.forceNoOverlap = forceNoOverlap

    @property
    def forceCylinder(self):
        return self._forceCylinder

    @forceCylinder.setter
    def forceCylinder(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.forceCylinder should be a bool, not a {type(value)}")
        self._forceCylinder = value

    @property
    def newSplitPlane(self):
        return self._newSplitPlane

    @newSplitPlane.setter
    def newSplitPlane(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.newSplitPlane should be a bool, not a {type(value)}")
        self._newSplitPlane = value

    @property
    def delLastNumber(self):
        return self._delLastNumber

    @delLastNumber.setter
    def delLastNumber(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.delLastNumber should be a bool, not a {type(value)}")
        self._delLastNumber = value

    @property
    def enlargeBox(self):
        return self._enlargeBox

    @enlargeBox.setter
    def enlargeBox(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError(f"geouned.Options.enlargeBox should be a Real, not a {type(value)}")
        if value < 0:
            raise ValueError(f"geouned.Options.enlargeBox should be above 0, not {value}")
        self._enlargeBox = value

    @property
    def nPlaneReverse(self):
        return self._nPlaneReverse

    @nPlaneReverse.setter
    def nPlaneReverse(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f"geouned.Options.nPlaneReverse should be a int, not a {type(value)}")
        self._nPlaneReverse = value

    @property
    def splitTolerance(self):
        return self._splitTolerance

    @splitTolerance.setter
    def splitTolerance(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError(f"geouned.Options.splitTolerance should be a Real, not a {type(value)}")
        if value < 0:
            raise ValueError(f"geouned.Options.splitTolerance should be above 0, not {value}")
        self._splitTolerance = value

    @property
    def scaleUp(self):
        return self._scaleUp

    @scaleUp.setter
    def scaleUp(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.scaleUp should be a bool, not a {type(value)}")
        self._scaleUp = value

    @property
    def quadricPY(self):
        return self._quadricPY

    @quadricPY.setter
    def quadricPY(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.quadricPY should be a bool, not a {type(value)}")
        self._quadricPY = value

    @property
    def Facets(self):
        return self._Facets

    @Facets.setter
    def Facets(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.Facets should be a bool, not a {type(value)}")
        self._Facets = value

    @property
    def prnt3PPlane(self):
        return self._prnt3PPlane

    @prnt3PPlane.setter
    def prnt3PPlane(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.prnt3PPlane should be a bool, not a {type(value)}")
        self._prnt3PPlane = value

    @property
    def forceNoOverlap(self):
        return self._forceNoOverlap

    @forceNoOverlap.setter
    def forceNoOverlap(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Options.forceNoOverlap should be a bool, not a {type(value)}")
        self._forceNoOverlap = value


class Tolerances:
    """A class for containing tolerances values

    Args:
        relativeTol (bool, optional): _description_. Defaults to False.
        relativePrecision (float, optional): relative precision. Defaults to 1.0e-6.
        value (float, optional): Tolerance in single value comparison. Defaults to 1.0e-6.
        distance (float, optional): General Distance Tolerance. Defaults to 1.0e-4.
        angle (float, optional): General Angle Tolerance. Defaults to 1.0e-4.
        pln_distance (float, optional): distance between planes equal planes if distance between parallel planes < 1e-4 cm. Defaults to 1.0e-4.
        pln_angle (float, optional): angle between axis. 1e-4 : planes separate each other 0.1mm each 1m. Defaults to 1.0e-4.
        cyl_distance (float, optional): distance between radius/center. Defaults to 1.0e-4.
        cyl_angle (float, optional): angle between axis. Defaults to 1.0e-4.
        sph_distance (float, optional): distance between radius/center. Defaults to 1.0e-4.
        kne_distance (float, optional): distance between apex. Defaults to 1.0e-4.
        kne_angle (float, optional): angle between semiangles/axis. Defaults to 1.0e-4.
        tor_distance (float, optional): distance between Major/Minor radii/center. Defaults to 1.0e-4.
        tor_angle (float, optional): angle between axis. Defaults to 1.0e-4.
        min_area (float, optional): minimum face area to consider in cell definition. Defaults to 1.0e-2.
    """

    def __init__(
        self,
        relativeTol: bool = False,
        relativePrecision: float = 1.0e-6,
        value: float = 1.0e-6,
        distance: float = 1.0e-4,
        angle: float = 1.0e-4,
        pln_distance: float = 1.0e-4,
        pln_angle: float = 1.0e-4,
        cyl_distance: float = 1.0e-4,
        cyl_angle: float = 1.0e-4,
        sph_distance: float = 1.0e-4,
        kne_distance: float = 1.0e-4,
        kne_angle: float = 1.0e-4,
        tor_distance: float = 1.0e-4,
        tor_angle: float = 1.0e-4,
        min_area: float = 1.0e-2,
    ):

        self.relativeTol = relativeTol
        self.relativePrecision = relativePrecision
        self.value = value
        self.distance = distance
        self.angle = angle
        self.pln_distance = pln_distance
        self.pln_angle = pln_angle
        self.cyl_distance = cyl_distance
        self.cyl_angle = cyl_angle
        self.sph_distance = sph_distance
        self.kne_distance = kne_distance
        self.kne_angle = kne_angle
        self.tor_distance = tor_distance
        self.tor_angle = tor_angle
        self.min_area = min_area

    @property
    def relativeTol(self):
        return self._relativeTol

    @relativeTol.setter
    def relativeTol(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"geouned.Tolerances.relativeTol should be a bool, not a {type(value)}")
        self._relativeTol = value

    @property
    def relativePrecision(self):
        return self._relativePrecision

    @relativePrecision.setter
    def relativePrecision(self, value: float):
        if not isinstance(value, float):
            raise TypeError(f"geouned.Tolerances.relativePrecision should be a float, not a {type(value)}")
        self._relativePrecision = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value: float):
        if not isinstance(value, float):
            raise TypeError(f"geouned.Tolerances.value should be a float, not a {type(value)}")
        self._value = value

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance: float):
        if not isinstance(distance, float):
            raise TypeError(f"geouned.Tolerances.distance should be a float, not a {type(distance)}")
        self._distance = distance

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, angle: float):
        if not isinstance(angle, float):
            raise TypeError(f"geouned.Tolerances.angle should be a float, not a {type(angle)}")
        self._angle = angle

    @property
    def pln_distance(self):
        return self._pln_distance

    @pln_distance.setter
    def pln_distance(self, pln_distance: float):
        if not isinstance(pln_distance, float):
            raise TypeError(f"geouned.Tolerances.pln_distance should be a float, not a {type(pln_distance)}")
        self._pln_distance = pln_distance

    @property
    def cyl_distance(self):
        return self._cyl_distance

    @cyl_distance.setter
    def cyl_distance(self, cyl_distance: float):
        if not isinstance(cyl_distance, float):
            raise TypeError(f"geouned.Tolerances.cyl_distance should be a float, not a {type(cyl_distance)}")
        self._cyl_distance = cyl_distance

    @property
    def cyl_angle(self):
        return self._cyl_angle

    @cyl_angle.setter
    def cyl_angle(self, cyl_angle: float):
        if not isinstance(cyl_angle, float):
            raise TypeError(f"geouned.Tolerances.cyl_angle should be a float, not a {type(cyl_angle)}")
        self._cyl_angle = cyl_angle

    @property
    def sph_distance(self):
        return self._sph_distance

    @sph_distance.setter
    def sph_distance(self, sph_distance: float):
        if not isinstance(sph_distance, float):
            raise TypeError(f"geouned.Tolerances.sph_distance should be a float, not a {type(sph_distance)}")
        self._sph_distance = sph_distance

    @property
    def pln_angle(self):
        return self._pln_angle

    @pln_angle.setter
    def pln_angle(self, pln_angle: float):
        if not isinstance(pln_angle, float):
            raise TypeError(f"geouned.Tolerances.pln_angle should be a float, not a {type(pln_angle)}")
        self._pln_angle = pln_angle

    @property
    def kne_distance(self):
        return self._kne_distance

    @kne_distance.setter
    def kne_distance(self, kne_distance: float):
        if not isinstance(kne_distance, float):
            raise TypeError(f"geouned.Tolerances.kne_distance should be a float, not a {type(kne_distance)}")
        self._kne_distance = kne_distance

    @property
    def kne_angle(self):
        return self._kne_angle

    @kne_angle.setter
    def kne_angle(self, kne_angle: float):
        if not isinstance(kne_angle, float):
            raise TypeError(f"geouned.Tolerances.kne_angle should be a float, not a {type(kne_angle)}")
        self._kne_angle = kne_angle

    @property
    def tor_distance(self):
        return self._tor_distance

    @tor_distance.setter
    def tor_distance(self, tor_distance: float):
        if not isinstance(tor_distance, float):
            raise TypeError(f"geouned.Tolerances.tor_distance should be a float, not a {type(tor_distance)}")
        self._tor_distance = tor_distance

    @property
    def tor_angle(self):
        return self._tor_angle

    @tor_angle.setter
    def tor_angle(self, tor_angle: float):
        if not isinstance(tor_angle, float):
            raise TypeError(f"geouned.Tolerances.tor_angle should be a float, not a {type(tor_angle)}")
        self._tor_angle = tor_angle

    @property
    def min_area(self):
        return self._min_area

    @min_area.setter
    def min_area(self, min_area: float):
        if not isinstance(min_area, float):
            raise TypeError(f"geouned.Tolerances.min_area should be a float, not a {type(min_area)}")
        self._min_area = min_area


class NumericFormat:
    """Numerical format options for each of the surface types.

    Args:
        P_abc (str, optional): Plane general a,b,c params. Defaults to "14.7e".
        P_d (str, optional): Plane general d params. Defaults to "14.7e".
        P_xyz (str, optional): PX/PY/PZ params. Defaults to "14.7e".
        S_r (str, optional): SO/SX/SY/SZ/S radius. Defaults to "14.7e".
        S_xyz (str, optional): SO/SX/SY/SZ/S center. Defaults to "14.7e".
        C_r (str, optional): Cylinder radius. Defaults to "12f".
        C_xyz (str, optional): Cylinder center. Defaults to "12f".
        K_xyz (str, optional): Cone apex. Defaults to "13.6e".
        K_tan2 (str, optional): Cone tan^2 value. Defaults to "13.6e".
        T_r (str, optional): Torus radii. Defaults to "14.7e".
        T_xyz (str, optional): Torus center. Defaults to "14.7e".
        GQ_1to6 (str, optional): GQ 1 to 6 coefficients (order 2 x2,y2,z2,xy,...). Defaults to "18.15f".
        GQ_7to9 (str, optional): GQ 7 to 9 coefficients (order 1 x,y,z). Defaults to "18.15f".
        GQ_10 (str, optional): GQ 10 coefficient. Defaults to "18.15f".
    """

    def __init__(
        self,
        P_abc: str = "14.7e",
        P_d: str = "14.7e",
        P_xyz: str = "14.7e",
        S_r: str = "14.7e",
        S_xyz: str = "14.7e",
        C_r: str = "12f",
        C_xyz: str = "12f",
        K_xyz: str = "13.6e",
        K_tan2: str = "13.6e",
        T_r: str = "14.7e",
        T_xyz: str = "14.7e",
        GQ_1to6: str = "18.15f",
        GQ_7to9: str = "18.15f",
        GQ_10: str = "18.15f",
    ):

        self.P_abc = P_abc
        self.P_d = P_d
        self.P_xyz = P_xyz
        self.S_r = S_r
        self.S_xyz = S_xyz
        self.C_r = C_r
        self.C_xyz = C_xyz
        self.K_xyz = K_xyz
        self.K_tan2 = K_tan2
        self.T_r = T_r
        self.T_xyz = T_xyz
        self.GQ_1to6 = GQ_1to6
        self.GQ_7to9 = GQ_7to9
        self.GQ_10 = GQ_10

    @property
    def P_abc(self):
        return self._P_abc

    @P_abc.setter
    def P_abc(self, P_abc: str):
        if not isinstance(P_abc, str):
            raise TypeError(f"geouned.Tolerances.P_abc should be a str, not a {type(P_abc)}")
        self._P_abc = P_abc

    @property
    def P_d(self):
        return self._P_d

    @P_d.setter
    def P_d(self, P_d: str):
        if not isinstance(P_d, str):
            raise TypeError(f"geouned.Tolerances.P_d should be a str, not a {type(P_d)}")
        self._P_d = P_d

    @property
    def P_xyz(self):
        return self._P_xyz

    @P_xyz.setter
    def P_xyz(self, P_xyz: str):
        if not isinstance(P_xyz, str):
            raise TypeError(f"geouned.Tolerances.P_xyz should be a str, not a {type(P_xyz)}")
        self._P_xyz = P_xyz

    @property
    def S_r(self):
        return self._S_r

    @S_r.setter
    def S_r(self, S_r: str):
        if not isinstance(S_r, str):
            raise TypeError(f"geouned.Tolerances.S_r should be a str, not a {type(S_r)}")
        self._S_r = S_r

    @property
    def S_xyz(self):
        return self._S_xyz

    @S_xyz.setter
    def S_xyz(self, S_xyz: str):
        if not isinstance(S_xyz, str):
            raise TypeError(f"geouned.Tolerances.S_xyz should be a str, not a {type(S_xyz)}")
        self._S_xyz = S_xyz

    @property
    def C_r(self):
        return self._C_r

    @C_r.setter
    def C_r(self, C_r: str):
        if not isinstance(C_r, str):
            raise TypeError(f"geouned.Tolerances.C_r should be a str, not a {type(C_r)}")
        self._C_r = C_r

    @property
    def C_xyz(self):
        return self._C_xyz

    @C_xyz.setter
    def C_xyz(self, C_xyz: str):
        if not isinstance(C_xyz, str):
            raise TypeError(f"geouned.Tolerances.C_xyz should be a str, not a {type(C_xyz)}")
        self._C_xyz = C_xyz

    @property
    def K_xyz(self):
        return self._K_xyz

    @K_xyz.setter
    def K_xyz(self, K_xyz: str):
        if not isinstance(K_xyz, str):
            raise TypeError(f"geouned.Tolerances.K_xyz should be a str, not a {type(K_xyz)}")
        self._K_xyz = K_xyz

    @property
    def K_tan2(self):
        return self._K_tan2

    @K_tan2.setter
    def K_tan2(self, K_tan2: str):
        if not isinstance(K_tan2, str):
            raise TypeError(f"geouned.Tolerances.K_tan2 should be a str, not a {type(K_tan2)}")
        self._K_tan2 = K_tan2

    @property
    def T_r(self):
        return self._T_r

    @T_r.setter
    def T_r(self, T_r: str):
        if not isinstance(T_r, str):
            raise TypeError(f"geouned.Tolerances.T_r should be a str, not a {type(T_r)}")
        self._T_r = T_r

    @property
    def T_xyz(self):
        return self._T_xyz

    @T_xyz.setter
    def T_xyz(self, T_xyz: str):
        if not isinstance(T_xyz, str):
            raise TypeError(f"geouned.Tolerances.T_xyz should be a str, not a {type(T_xyz)}")
        self._T_xyz = T_xyz

    @property
    def GQ_1to6(self):
        return self._GQ_1to6

    @GQ_1to6.setter
    def GQ_1to6(self, GQ_1to6: str):
        if not isinstance(GQ_1to6, str):
            raise TypeError(f"geouned.Tolerances.GQ_1to6 should be a str, not a {type(GQ_1to6)}")
        self._GQ_1to6 = GQ_1to6

    @property
    def GQ_7to9(self):
        return self._GQ_7to9

    @GQ_7to9.setter
    def GQ_7to9(self, GQ_7to9: str):
        if not isinstance(GQ_7to9, str):
            raise TypeError(f"geouned.Tolerances.GQ_7to9 should be a str, not a {type(GQ_7to9)}")
        self._GQ_7to9 = GQ_7to9

    @property
    def GQ_10(self):
        return self._GQ_10

    @GQ_10.setter
    def GQ_10(self, GQ_10: str):
        if not isinstance(GQ_10, str):
            raise TypeError(f"geouned.Tolerances.GQ_10 should be a str, not a {type(GQ_10)}")
        self._GQ_10 = GQ_10


class Settings:
    """Settings for changing the way the CAD to CSG conversion is done

    Args:
        matFile (str, optional): _description_. Defaults to "".
        voidGen (bool, optional): Generate voids of the geometry. Defaults
            to True.
        debug (bool, optional): Write step files of original and decomposed
            solids, for each solid in the STEP file. Defaults to False.
        compSolids (bool, optional): Join subsolids of STEP file as a single
            compound solid. Step files generated with SpaceClaim have not
            exactly the same level of solids as FreeCAD. It may a happened
            that solids defined has separated solids are read by FreeCAD
            as a single compound solid (and will produce only one MCNP
            cell). In this case compSolids should be set to False. Defaults
            to False.
        simplify (str, optional): Simplify the cell definition considering
            relative surfaces position and using Boolean logics. Available
            options are: "no" no optimization, "void" only void cells are
            simplified. Algorithm is faster but the simplification is not
            optimal. "voidfull" : only void cells are simplified with the
            most optimal algorithm. The time of the conversion can be
            multiplied by 5 or more. "full" : all the cells (solids and
            voids) are simplified. Defaults to "No".
        exportSolids (str, optional): Export CAD solid after reading.
            The execution is stopped after export, the translation is not
            carried out. Defaults to "".
        minVoidSize (float, optional): Minimum size of the edges of the
            void cell. Units are in mm. Defaults to 200.0.
        maxSurf (int, optional): #TODO
        maxBracket (int, optional): Maximum number of brackets (solid
            complementary) allowed in void cell definition. Defaults to 30.
        voidMat (list, optional): Assign a material defined by the user
            instead of void for cells without material definition and the
            cells generated in the automatic void generation. The format
            is a 3 valued tuple (mat_label, mat_density, mat_description).
            Example (100,1e-3,'Air assigned to Void'). Defaults to [].
        voidExclude (list, optional): #TODO see issue 87. Defaults to [].
        startCell (int, optional): Starting cell numbering label. Defaults to 1.
        startSurf (int, optional): Starting surface numbering label. Defaults to 1.
        sort_enclosure (bool, optional): If enclosures are defined in the
            CAD models, the voids cells of the enclosure will be located in
            the output file in the same location where the enclosure solid
            is located in the CAD solid tree.. Defaults to False.
    """

    def __init__(
        self,
        matFile: str = "",
        voidGen: bool = True,
        debug: bool = False,
        compSolids: bool = False,
        simplify: str = "no",
        exportSolids: typing.Optional[str] = None,
        minVoidSize: float = 200.0,  # units mm
        maxSurf: int = 50,
        maxBracket: int = 30,
        voidMat: list = [],
        voidExclude: list = [],
        startCell: int = 1,
        startSurf: int = 1,
        sort_enclosure: bool = False,
    ):

        self.matFile = matFile
        self.voidGen = voidGen
        self.debug = debug
        self.compSolids = compSolids
        self.simplify = simplify
        self.exportSolids = exportSolids
        self.minVoidSize = minVoidSize
        self.maxSurf = maxSurf
        self.maxBracket = maxBracket
        self.voidMat = voidMat
        self.voidExclude = voidExclude
        self.startCell = startCell
        self.startSurf = startSurf
        self.sort_enclosure = sort_enclosure

    @property
    def matFile(self):
        return self._matFile

    @matFile.setter
    def matFile(self, matFile: str):
        if not isinstance(matFile, str):
            raise TypeError(f"geouned.Settings.matFile should be a str, not a {type(matFile)}")
        self._matFile = matFile

    @property
    def voidGen(self):
        return self._voidGen

    @voidGen.setter
    def voidGen(self, voidGen: bool):
        if not isinstance(voidGen, bool):
            raise TypeError(f"geouned.Settings.voidGen should be a bool, not a {type(voidGen)}")
        self._voidGen = voidGen

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, debug: bool):
        if not isinstance(debug, bool):
            raise TypeError(f"geouned.Settings.debug should be a bool, not a {type(debug)}")
        self._debug = debug

    @property
    def compSolids(self):
        return self._compSolids

    @compSolids.setter
    def compSolids(self, compSolids: bool):
        if not isinstance(compSolids, bool):
            raise TypeError(f"geouned.Settings.compSolids should be a bool, not a {type(compSolids)}")
        self._compSolids = compSolids

    @property
    def simplify(self):
        return self._simplify

    @simplify.setter
    def simplify(self, simplify: str):
        if not isinstance(simplify, str):
            raise TypeError(f"geouned.Settings.simplify should be a str, not a {type(simplify)}")
        self._simplify = simplify

    @property
    def exportSolids(self):
        return self._exportSolids

    @exportSolids.setter
    def exportSolids(self, exportSolids: str):
        if exportSolids == None:
            pass
        elif not isinstance(exportSolids, str):
            raise TypeError(f"geouned.Settings.exportSolids should be a str, not a {type(exportSolids)}")
        self._exportSolids = exportSolids

    @property
    def minVoidSize(self):
        return self._minVoidSize

    @minVoidSize.setter
    def minVoidSize(self, minVoidSize: float):
        if not isinstance(minVoidSize, float):
            raise TypeError(f"geouned.Settings.minVoidSize should be a float, not a {type(minVoidSize)}")
        self._minVoidSize = minVoidSize

    @property
    def maxSurf(self):
        return self._maxSurf

    @maxSurf.setter
    def maxSurf(self, maxSurf: int):
        if not isinstance(maxSurf, int):
            raise TypeError(f"geouned.Settings.maxSurf should be a int, not a {type(maxSurf)}")
        self._maxSurf = maxSurf

    @property
    def maxBracket(self):
        return self._maxBracket

    @maxBracket.setter
    def maxBracket(self, maxBracket: int):
        if not isinstance(maxBracket, int):
            raise TypeError(f"geouned.Settings.maxBracket should be a int, not a {type(maxBracket)}")
        self._maxBracket = maxBracket

    @property
    def voidMat(self):
        return self._voidMat

    @voidMat.setter
    def voidMat(self, voidMat: list):
        if not isinstance(voidMat, list):
            raise TypeError(f"geouned.Settings.voidMat should be a list, not a {type(voidMat)}")
        if len(voidMat) == 3:
            entry0, entry1, entry2 = voidMat
            if not isinstance(entry0, int):
                raise TypeError(f"first entry of geouned.Settings.voidMat should be an int, not a {type(entry0)}")
            if not isinstance(entry1, int):
                if not isinstance(entry1, float):
                    raise TypeError(f"second entry of geouned.Settings.voidMat should be an int or float, not a {type(entry1)}")
            if not isinstance(entry2, str):
                raise TypeError(f"third entry of geouned.Settings.voidMat should be a str, not a {type(entry2)}")
        elif len(voidMat) > 0:
            raise TypeError(f"geouned.Settings.voidMat should be a list with 3 elements or void list")
        self._voidMat = voidMat

    @property
    def voidExclude(self):
        return self._voidExclude

    @voidExclude.setter
    def voidExclude(self, voidExclude: list):
        if not isinstance(voidExclude, list):
            raise TypeError(f"geouned.Settings.voidExclude should be a list, not a {type(voidExclude)}")
        for entry in voidExclude:
            if not isinstance(entry, int):
                raise TypeError(f"geouned.Settings.voidExclude should be a list of ints, not a {type(entry)}")
        self._voidExclude = voidExclude

    @property
    def startCell(self):
        return self._startCell

    @startCell.setter
    def startCell(self, startCell: int):
        if not isinstance(startCell, int):
            raise TypeError(f"geouned.Settings.startCell should be a int, not a {type(startCell)}")
        self._startCell = startCell

    @property
    def startSurf(self):
        return self._startSurf

    @startSurf.setter
    def startSurf(self, startSurf: int):
        if not isinstance(startSurf, int):
            raise TypeError(f"geouned.Settings.startSurf should be a int, not a {type(startSurf)}")
        self._startSurf = startSurf

    @property
    def sort_enclosure(self):
        return self._sort_enclosure

    @sort_enclosure.setter
    def sort_enclosure(self, sort_enclosure: bool):
        if not isinstance(sort_enclosure, bool):
            raise TypeError(f"geouned.Settings.sort_enclosure should be a bool, not a {type(sort_enclosure)}")
        self._sort_enclosure = sort_enclosure



""" Modification from 0.9.5 
  - volInput card name substituted by volSDEF card name
  - new card volCARD (values : True/False) used to insert the CAD cell volume into the cell definition
    using the MCNP card VOL. Default value True  
  - Fixed bug. Code crashing when automatic void calculation was not enabled.
  - Fixed bug. Comment line might not be in the correct order following de CAD tree labels

Modification in 0.9.6 Release Date 24/11/2022
  - Small change in the reading of material file.
  - In "CellDefinition.py" file. "is_opposite" function is call with rhe correct tolerance value instead of default value of the is_opposite function.  

Modification in 0.9.7 Release Date 06/12/2022
  - Compatibility with FreeCAD version higher than 0.18.
  - fixed bug with sphere Objects

Modification in 0.9.7 Release Date 06/12/2022
  - Removed external packages Boolean. Subtituted by own make package
  - Full cell optimization using CTable approach and own factorization function
  - Cylinders instead cones can be used  as auxillary surface for Torus surface definition 
  - fix bugs several additional planes for the same reversed surface
  - new module for plane splitting in decomposition modules. All planes are considerer equivalent (PX,PY,PZ,P), parallel planes are grouped together. 
    Decomposition is performed by splitting first with group of parallel planes with lowest number of elements. 
    Old version module is still available by switching the option keyword "newSplitPlane" to False. 
  - New option added when enclosure are used. The option "sort_enclosure" True (Default False) will produce output file where solids and void of an enclosure
    are grouped together (first solid cell of the enclosure, then voids)
  - Options has been added to enlarge Box dimension when Constraint table is evaluated
 
 Modification in 0.9.8 
  - fix bug with two cylinders/cones intersection
  - Introduce splitting planes to cut rounded edges(edge made with cylinder)
  - fixed bug with torus (I hope so!)
  - new card "dummyMat" to write dummy material card (associated to material labels present in cell definition) in input file.

 new in version 0.9.8.2
  - Javi modification for memory bug for geomtry with large plane numbers
  - small changes in decomposition process
  - new verbose option to remove some output on screen

 new in version 1.0
  - allow to convert geometry to openMC XML format and script input file
  - change the entry "mcnpfile" by "geometryName" in FILES section. Now geometryName is the generic name of output geometry file.
    mncp input file will have ".mcnp" extension and openMC XML the ".xml" extension.
  - new entry "outFormat" is added in FILES section. "outFormat" is used to select the output geometry format (mcnp and/or openMC_xml). 
    Default value is only mcnp
  - new entry splitTolerance. Change the Tolerance value of the FreeCAD function split. Default Value  0.
  - new entry Verbose. Print on screen warning. Default False.

 new in version 1.0.1
  - implement new conversion module when all cells are made of triangular plane faces
  - bug fixed void complementary of level1 'AND' sequence
  - changes in simplify function of BoolSequence class
  - implementation of the geometry conversion to Serpent code input file (by A. Valentine UKAEA)
  - fixed bug with torodial surfaces.
  - implementation of new, more robust, cell definition, where adjacent solid cell definition is rested to current solid cell definition (forcing no overlaping cell).
  - fix bug additional plane with sphere
  - improvement of cell simplification method
  - fix bug. bad parameter normalization multiple output formats
  - Change tolerance values if BOPTsplit function crash
  - fix some bugs with Torus (not all)
  - implementation of the geometry conversion to PHITS code input file (by T. Okano Nissin High Voltage Corporation (NHVC))
"""



import string

version = "3.6"


def make_label(item):
    if 0 < len(item):
        return "{" + item + "}"
    else:
        return item


class SafeDict(dict):

    def __getitem__(self, item):
        return super(SafeDict, self).__getitem__(item) or ""

    def __missing__(self, key):
        return make_label(key)


# ' 1 #'
class PartialFormatter(string.Formatter):

    def format(*args, **kwargs):
        if not args:
            raise TypeError("descriptor 'format' of 'Formatter' object " "needs an argument")
        self, args = args[0], args[1:]  # allow the "self" keyword be passed
        try:
            format_string, args = (
                args[0],
                args[1:],
            )  # allow the "format_string" keyword be passed
        except IndexError:
            if "format_string" in kwargs:
                format_string = kwargs.pop("format_string")
            else:
                raise TypeError("format() missing 1 required positional " "argument: 'format_string'")
        return string.Formatter.vformat(self, format_string, args, SafeDict(kwargs))

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth, auto_arg_index=0):
        """Clone from Python3 version to fix Python2 mess with unnamed {} format specifiers"""
        if recursion_depth < 0:
            raise ValueError("Max string recursion exceeded")
        result = []
        for literal_text, field_name, format_spec, conversion in self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                # handle arg indexing when empty field_names are given.
                if field_name == "":
                    if auto_arg_index is False:
                        raise ValueError("cannot switch from manual field " "specification to automatic field " "numbering")
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError("cannot switch from manual field " "specification to automatic field " "numbering")
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                #  and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                if version == "2.7":
                    format_spec = self._vformat(
                        format_spec,
                        args,
                        kwargs,
                        used_args,
                        recursion_depth - 1,
                        auto_arg_index=auto_arg_index,
                    )
                else:
                    format_spec, auto_arg_index = self._vformat(
                        format_spec,
                        args,
                        kwargs,
                        used_args,
                        recursion_depth - 1,
                        auto_arg_index=auto_arg_index,
                    )
                # format the object and append to the result
                result.append(self.format_field(obj, format_spec))

        result = "".join(result)

        if version == "2.7":
            return result
        else:
            return result, auto_arg_index



#!/usr/bin/env python3

"""
Convert Cad geometry to CSG format for use in Monte Carlo Codes
"""

import argparse

import geouned

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-i",
    "--input",
    type=str,
    default="config.json",
    help="The path to the config JSON file",
)
args = parser.parse_args()


def main():
    geouned.CadToCsg.from_json(args.input)


if __name__ == "__main__":
    main()
